#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""
Debusine integration tests.

Test mmdebstrap related code.
"""

import unittest
from typing import override

from utils.common import Configuration
from utils.integration_test_helpers_mixin import IntegrationTestHelpersMixin
from utils.server import DebusineServer


class IntegrationTaskMmDebstrapTests(
    IntegrationTestHelpersMixin, unittest.TestCase
):
    """
    Integration test for the mmdebstrap task.

    These tests assume:
    - debusine-server is running
    - debusine-worker is running (connected to the server)
    - debusine-client is correctly configured
    """

    TASK_NAME = "mmdebstrap"

    @override
    def setUp(self) -> None:
        super().setUp()
        # If debusine-server or nginx was launched just before the
        # integration-tests-task-mmdebstrap.py is launched the debusine-server
        # might not be yet available. Let's wait for the debusine-server to be
        # reachable if it's not ready yet
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )

    def test_mmdebstrap(self) -> None:
        """Create a workflow to run mmdebstrap."""
        # This asserts that the workflow succeeds.
        self.create_system_images_mmdebstrap(["bookworm", "trixie"])



# Generated by Django 4.2.23 on 2026-02-04 17:30

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0041_add_group_workspace_namespace'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='scheduler_tags_provided',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255),
                blank=True,
                default=list,
                help_text='final set of task-provided tags used for scheduling',
                size=None,
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='scheduler_tags_required',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255),
                blank=True,
                default=list,
                help_text='final set of task-required tags used for scheduling',
                size=None,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: work requests."""

import itertools
import logging
from typing import Any, override

import pydantic
from django.http import Http404
from django.http.response import HttpResponseBase
from rest_framework import status
from rest_framework.exceptions import (
    NotFound,
    PermissionDenied,
    ValidationError,
)
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    NotificationChannel,
    WorkRequest,
)
from debusine.db.models.permissions import anonymous_access
from debusine.db.models.work_requests import (
    CannotAbort,
    CannotRetry,
    CannotUnblock,
    WorkRequestQuerySet,
)
from debusine.db.tasks import DBTask
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    WorkRequestConfirmSerializer,
    WorkRequestExternalDebsignRequestSerializer,
    WorkRequestExternalDebsignSerializer,
    WorkRequestSerializer,
    WorkRequestSerializerWithConfiguredTaskData,
    WorkRequestUnblockSerializer,
    WorkRequestUpdateSerializer,
)
from debusine.server.tasks.wait import Confirm
from debusine.server.tasks.wait.models import ExternalDebsignDynamicData
from debusine.server.views.base import (
    BaseAPIView,
    CanDisplayFilterBackend,
    GenericAPIViewBase,
    ListAPIViewBase,
    RetrieveAPIViewBase,
    UpdateAPIViewBase,
)
from debusine.server.views.rest import (
    IsGet,
    IsUserTokenAuthenticated,
    IsWorkRequestTokenAuthenticated,
)
from debusine.tasks import TaskConfigError, inputs
from debusine.tasks.models import ActionTypes

logger = logging.getLogger(__name__)


class WorkRequestPagination(CursorPagination):
    """Pagination for lists of work requests."""

    ordering = "-created_at"


class WorkRequestMixin(GenericAPIViewBase[WorkRequest]):
    """Look up work requests for later use with WorkRequestSerializer."""

    @override
    def get_queryset(self) -> WorkRequestQuerySet[Any]:
        """Get the query set for this view."""
        return (
            WorkRequest.objects.in_current_scope()
            .select_related(
                "aborted_by",
                "created_by",
                "parent__workspace__scope",
                "workflow_root__workspace__scope",
                "workspace__scope",
            )
            .can_display(context.pc)
        )

    @override
    def get_object(self) -> WorkRequest:
        """Override to return more API-friendly errors."""
        try:
            return super().get_object()
        except Http404 as exc:
            raise NotFound(str(exc))


class WorkRequestView(
    WorkRequestMixin,
    ListAPIViewBase[WorkRequest],
    UpdateAPIViewBase[WorkRequest],
    BaseAPIView,
):
    """View used by the debusine client to get information of a WorkRequest."""

    permission_classes = [
        IsUserTokenAuthenticated | (IsWorkRequestTokenAuthenticated & IsGet)
    ]
    serializer_class = WorkRequestSerializerWithConfiguredTaskData
    filter_backends = [CanDisplayFilterBackend]
    pagination_class = WorkRequestPagination
    lookup_url_kwarg = "work_request_id"

    @override
    def get(
        self, request: Request, work_request_id: int | None = None
    ) -> Response:
        """Return status information for WorkRequest or not found."""
        self.enforce(anonymous_access)  # We filter can_display
        if work_request_id is None:
            return super().get(request)

        return Response(
            self.serializer_class(self.get_object()).data,
            status=status.HTTP_200_OK,
        )

    def post(self, request: Request) -> Response:
        """Create a new work request."""
        data = request.data.copy()

        data["created_by"] = request.auth.user.id
        work_request_deserialized = self.serializer_class(
            data=data,
            only_fields=[
                'task_name',
                'task_data',
                'event_reactions',
                'workspace',
                'created_by',
            ],
        )

        try:
            work_request_deserialized.is_valid(raise_exception=True)
        except DebusineAPIException as e:
            logger.debug(
                "Error creating work request. Could not be deserialized: "
                "%s (detail: %s, validation errors: %s)",
                e.debusine_title,
                e.debusine_detail,
                e.debusine_validation_errors,
            )
            raise

        workspace = work_request_deserialized.validated_data["workspace"]
        self.set_current_workspace(workspace)
        self.enforce(workspace.can_create_work_requests)

        self._validate_event_reactions(
            work_request_deserialized.validated_data.get(
                "event_reactions_json", {}
            )
        )

        if len(non_existing_channels := self._non_existing_channels(data)) > 0:
            raise DebusineAPIException(
                title=f"Non-existing channels: {sorted(non_existing_channels)}"
            )

        task_name = data['task_name']
        task_type = self._validate_task_name(
            task_name, request.auth.user.is_superuser
        )

        # Return an error if compute_dynamic_data() raise an exception
        # (otherwise it happens during the scheduler() and the user
        # doesn't know why)
        #
        # If more checks are added here: they might need to be added as well
        # in the view WorkRequestCreateView.form_valid()
        w = work_request_deserialized.save(task_type=task_type)

        try:
            task = w.get_task()
        except TaskConfigError as exc:
            exception: Exception = exc
            errors = None
            if isinstance(exc.original_exception, pydantic.ValidationError):
                exception = exc.original_exception
                errors = exc.original_exception.errors()

            raise DebusineAPIException(
                title='Cannot create task',
                detail=f"Task data: {w.task_data} Error: {exception!r}",
                pydantic_validation_errors=errors,
            )

        try:
            task.resolve_inputs(inputs.Stage.RUNNING)
        except Exception as exc:
            if isinstance(exc, AssertionError):
                logger.exception(
                    (
                        "Assertion failed resolving input fields"
                        " for task %s with data %s"
                    ),
                    task_name,
                    w.task_data,
                )
            raise DebusineAPIException(
                title=(
                    "Cannot create work request:"
                    " error resolving task input fields"
                ),
                detail=f"Task data: {w.task_data} Error: {exc!r}",
            )

        try:
            task.apply_task_configuration()
        except Exception as exc:
            if isinstance(exc, AssertionError):
                logger.exception(
                    (
                        "Assertion failed applying task configuration"
                        " for task %s with data %s"
                    ),
                    task_name,
                    w.task_data,
                )
            raise DebusineAPIException(
                title=(
                    "Cannot create work request:"
                    " error applying task configuration"
                ),
                detail=f"Task data: {w.task_data} Error: {exc!r}",
            )

        try:
            task.compute_dynamic_data()
        except Exception as exc:
            if isinstance(exc, AssertionError):
                logger.exception(
                    (
                        "Assertion failed computing dynamic data for task %s "
                        "with data %s"
                    ),
                    task_name,
                    w.task_data,
                )
            raise DebusineAPIException(
                title=(
                    "Cannot create work request: error computing dynamic data"
                ),
                detail=f"Task data: {w.task_data} Error: {exc!r}",
            )

        w.refresh_from_db()
        w.mark_pending()

        return Response(
            self.serializer_class(w).data, status=status.HTTP_200_OK
        )

    def _validate_event_reactions(
        self, event_reactions: dict[str, Any]
    ) -> None:
        """Only accept send-notification actions."""
        for reactions in event_reactions.values():
            for action in reactions:
                if action["action"] != ActionTypes.SEND_NOTIFICATION:
                    raise DebusineAPIException(
                        title="Invalid event_reactions",
                        detail=(
                            f"Action type {action['action']!r} is not allowed "
                            f"here"
                        ),
                    )

    def _validate_task_name(
        self, task_name: str, is_superuser: bool
    ) -> TaskTypes:
        """Validate that task_name is valid, return the task_type."""
        task_types = [TaskTypes.WORKER]

        if is_superuser:
            task_types.append(TaskTypes.SERVER)

        for task_type in task_types:
            if DBTask.is_valid_task_name(task_type, task_name):
                return task_type

        valid_task_names = ', '.join(
            itertools.chain(
                *(DBTask.task_names(task_type) for task_type in task_types),
            ),
        )
        logger.debug(
            "Error creating work request: task name is not registered: %s",
            task_name,
        )
        raise DebusineAPIException(
            title="Cannot create work request: task name is not registered",
            detail=(
                f'Task name: "{task_name}". '
                f"Registered task names: {valid_task_names}"
            ),
        )

    @staticmethod
    def _non_existing_channels(data: dict[str, Any]) -> set[str]:
        """Return non existing channels from event_reactions."""
        non_existing_channels: set[str] = set()

        for reactions in data.get("event_reactions", {}).values():
            for action in reactions:
                assert (
                    action["action"] == ActionTypes.SEND_NOTIFICATION
                )  # filtered above
                channel_name = action["channel"]
                try:
                    NotificationChannel.objects.get(name=channel_name)
                except NotificationChannel.DoesNotExist:
                    non_existing_channels.add(channel_name)

        return non_existing_channels

    @override
    def patch(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """Update properties of a work request."""
        # Only a subset of properties may be patched.
        work_request_deserialized = WorkRequestUpdateSerializer(
            data=request.data
        )
        work_request_deserialized.is_valid(raise_exception=True)
        work_request = self.get_object()

        # Only users with appropriate permissions may set priority
        # adjustments.
        priority_adjustment = work_request_deserialized.validated_data.get(
            "priority_adjustment"
        )
        if priority_adjustment is None:
            pass
        else:
            token = request.auth
            # permission_classes is declared such that we won't get this far
            # unless the request has an enabled token with an associated
            # user.
            assert token is not None
            assert token.user is not None
            # Users with manage_workrequest_priorities have full access.
            if token.user.has_perm("db.manage_workrequest_priorities"):
                pass
            # Users can set non-positive priority adjustments on their own
            # work requests.
            elif (
                priority_adjustment <= 0
                and token.user == work_request.created_by
            ):
                pass
            else:
                raise DebusineAPIException(
                    title="You are not permitted to set priority adjustments",
                    status_code=status.HTTP_403_FORBIDDEN,
                )

        # TODO: Replace with work_request.can_edit
        self.enforce(work_request.workspace.can_create_work_requests)

        return super().patch(request, *args, **kwargs)


class WorkRequestRetryView(
    WorkRequestMixin, GenericAPIViewBase[WorkRequest], BaseAPIView
):
    """View used by the debusine client to get information of a WorkRequest."""

    permission_classes = [IsUserTokenAuthenticated]
    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "work_request_id"

    def post(self, *args: Any, **kwargs: Any) -> Response:
        """Retry a work request."""
        work_request = self.get_object()
        # This duplicates a check in .retry()
        self.enforce(work_request.can_retry)

        try:
            new_work_request = work_request.retry()
        except CannotRetry as e:
            raise DebusineAPIException(
                title="Cannot retry work request",
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=str(e),
            )

        return Response(
            WorkRequestSerializer(new_work_request).data,
            status=status.HTTP_200_OK,
        )


class WorkRequestAbortView(
    WorkRequestMixin, GenericAPIViewBase[WorkRequest], BaseAPIView
):
    """View used to abort a work request."""

    permission_classes = [IsUserTokenAuthenticated]
    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "work_request_id"

    def post(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """Abort a work request."""
        work_request = self.get_object()
        # This duplicates a check in .abort()
        self.enforce(work_request.can_abort)

        try:
            work_request.abort()
        except CannotAbort as e:
            raise DebusineAPIException(
                title="Cannot abort work request",
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=str(e),
            )

        logger.info(
            "%s aborted work request %s", request.auth.user, work_request
        )

        work_request.refresh_from_db()
        return Response(
            WorkRequestSerializer(work_request).data, status=status.HTTP_200_OK
        )


class WorkRequestUnblockView(
    WorkRequestMixin, GenericAPIViewBase[WorkRequest], BaseAPIView
):
    """View used to unblock a work request awaiting manual approval."""

    permission_classes = [IsUserTokenAuthenticated]
    serializer_class = WorkRequestUnblockSerializer
    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "work_request_id"

    def post(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """Edit a manual unblock."""
        work_request = self.get_object()
        self.set_current_workspace(work_request.workspace)
        self.enforce(work_request.can_unblock)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        notes = serializer.validated_data.get("notes")
        action = serializer.validated_data.get("action")

        token = self.request.auth
        # permission_classes is declared such that we won't get this far
        # unless the request has an enabled token with an associated user.
        assert token is not None
        assert token.user is not None

        try:
            work_request.review_manual_unblock(
                user=token.user, notes=notes, action=action
            )
        except CannotUnblock as e:
            raise DebusineAPIException(title=str(e))

        return Response(
            WorkRequestSerializer(work_request).data, status=status.HTTP_200_OK
        )


class IsWorkRequestCreator(BasePermission):
    """Allow access if the requester created the work request."""

    @override
    def has_object_permission(
        self,
        request: Request,
        view: APIView,  # noqa: ARG002, U100
        obj: Any,
    ) -> bool:
        """Return True if the requester created the work request."""
        if isinstance(obj, WorkRequest) and request.auth.user != obj.created_by:
            # Raise an explicit exception in this case for the sake of a
            # better error message.
            raise PermissionDenied(
                "Only the user who created the work request can provide a "
                "signature"
            )

        return True


class WorkRequestExternalDebsignView(
    WorkRequestMixin, RetrieveAPIViewBase[WorkRequest], BaseAPIView
):
    """View used to handle `ExternalDebsign` wait tasks."""

    permission_classes = [
        IsUserTokenAuthenticated,
        IsGet | IsWorkRequestCreator,
    ]
    serializer_class = WorkRequestExternalDebsignSerializer
    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "work_request_id"

    @override
    def get_object(self) -> WorkRequest:
        """Override to add constraints and return more API-friendly errors."""
        self.enforce(anonymous_access)  # We filter to can_display()
        work_request = super().get_object()

        if (
            work_request.task_type != TaskTypes.WAIT
            or work_request.task_name != "externaldebsign"
        ):
            raise ValidationError(
                detail=(
                    f"Expected work request to be Wait/ExternalDebsign; got "
                    f"{work_request.task_type}/{work_request.task_name}"
                )
            )
        if work_request.status != WorkRequest.Statuses.RUNNING:
            raise ValidationError(
                detail=(
                    f"Work request is not running. "
                    f"Status: {work_request.status}"
                )
            )

        return work_request

    def post(
        self, request: Request, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Complete an `ExternalDebsign` task."""
        work_request = self.get_object()
        WorkRequest.objects.filter(
            id=work_request.id
        ).lock_workflows_for_update()
        user = context.user
        # This view's permission_classes ensure that the user can't be
        # anonymous by this point.
        assert user.is_authenticated
        serializer = WorkRequestExternalDebsignRequestSerializer(
            data=request.data
        )
        serializer.is_valid(raise_exception=True)

        dynamic_task_data = ExternalDebsignDynamicData(
            **self.get_serializer(work_request).data["dynamic_task_data"]
        )
        unsigned = Artifact.objects.get(id=dynamic_task_data.unsigned_id)
        signed = serializer.validated_data["signed_artifact"]

        unsigned_files = {
            fia.path: fia.file.hash_digest
            for fia in unsigned.fileinartifact_set.select_related("file")
        }
        signed_files = {
            fia.path: fia.file.hash_digest
            for fia in signed.fileinartifact_set.select_related("file")
        }

        new_paths = set(signed_files) - set(unsigned_files)
        if new_paths:
            raise ValidationError(
                detail=(
                    f"Signed upload adds extra files when compared to "
                    f"unsigned upload. New paths: {sorted(new_paths)}"
                )
            )

        changed_paths = {
            path
            for path, hash_digest in signed_files.items()
            if unsigned_files.get(path) != hash_digest
        }
        if not any(path.endswith(".changes") for path in changed_paths):
            raise ValidationError(
                detail="Signed upload does not change the .changes file"
            )
        if not all(
            path.endswith((".changes", ".dsc", ".buildinfo"))
            for path in changed_paths
        ):
            raise ValidationError(
                detail=(
                    f"Signed upload changes more than "
                    f".changes/.dsc/.buildinfo. "
                    f"Changed paths: {sorted(changed_paths)}"
                )
            )

        ArtifactRelation.objects.create(
            artifact=signed,
            target=unsigned,
            type=ArtifactRelation.Relations.RELATES_TO,
        )
        signed.created_by_work_request = work_request
        signed.save()
        work_request.mark_completed(WorkRequest.Results.SUCCESS, user=user)

        return Response(self.get_serializer(work_request).data)


class WorkRequestConfirmView(
    WorkRequestMixin, RetrieveAPIViewBase[WorkRequest], BaseAPIView
):
    """View used to handle ``Confirm`` wait tasks."""

    permission_classes = [IsUserTokenAuthenticated]
    serializer_class = WorkRequestConfirmSerializer
    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "work_request_id"

    @override
    def get_object(self) -> WorkRequest:
        work_request = super().get_object()

        if (
            work_request.task_type != TaskTypes.WAIT
            or work_request.task_name != "confirm"
        ):
            raise ValidationError(
                detail=(
                    f"Expected work request to be Wait/Confirm; got "
                    f"{work_request.task_type}/{work_request.task_name}"
                )
            )
        if work_request.status != WorkRequest.Statuses.RUNNING:
            raise ValidationError(
                detail=(
                    f"Work request is not running. "
                    f"Status: {work_request.status}"
                )
            )

        return work_request

    def post(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """Confirm or deny a work request."""
        work_request = self.get_object()
        self.set_current_workspace(work_request.workspace)
        self.enforce(work_request.can_confirm)
        task = work_request.get_task()
        assert isinstance(task, Confirm)
        user = context.user
        # Enforcing `can_confirm` ensures that the user can't be anonymous
        # by this point.
        assert user.is_authenticated
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        task.respond(
            user=user,
            confirm=serializer.validated_data["confirm"],
            comment=serializer.validated_data.get("comment"),
        )
        work_request.refresh_from_db()

        return Response(
            WorkRequestSerializerWithConfiguredTaskData(work_request).data,
            status=status.HTTP_200_OK,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test the Worker Pool base API."""

from debusine.server.worker_pools import DummyWorkerPool, provider_interface
from debusine.test.django import TestCase


class TestWorkerPoolInterface(TestCase):
    """Tests for WorkerPoolInterface."""

    def test_lookup_test(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool_api = provider_interface(worker_pool)
        self.assertIsInstance(worker_pool_api, DummyWorkerPool)
        self.assertEqual(worker_pool_api.worker_pool, worker_pool)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine command line interface, suite management commands.

These are "porcelain": they implement opinionated operations on
:collection:`debian:suite` collections and their contents, rather than basic
"plumbing" operations.
"""

import argparse
from collections.abc import Generator
from typing import Any, override

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import WorkspaceCommand
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionDataNew,
    WorkflowTemplateData,
    WorkflowTemplateDataNew,
)


class Create(WorkspaceCommand, group="archive.suite"):
    """Create a suite in an archive."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        super().configure(parser)
        parser.add_argument("suite", type=str, help="Suite name")
        parser.add_argument(
            "--component",
            type=str,
            dest="components",
            action="append",
            help=(
                "Component name to add to the new suite; may be given more "
                "than once (default: [main])"
            ),
        )
        # Although this is currently required, it's an option because we
        # might be able to infer it in future (e.g. from a base suite).
        parser.add_argument(
            "--architecture",
            type=str,
            required=True,
            dest="architectures",
            action="append",
            help=(
                "Architecture name to add to the new suite; may be given more "
                "than once"
            ),
        )
        parser.add_argument(
            "--base-workflow-template",
            type=str,
            help=(
                "Create a workflow template to publish to this suite, using "
                "this one as a base; may be from this workspace or a parent"
            ),
        )

    def iter_workspace_names(self) -> Generator[str]:
        """Iterate over names of this workspace and its parents."""
        yield self.workspace
        with self._api_call_or_fail():
            inheritance = self.debusine.get_workspace_inheritance(
                self.workspace
            )
        for parent in inheritance.chain:
            # Always set in API responses.
            assert parent.workspace is not None
            # We currently only support looking for a base workflow template
            # in the current scope, because our client may not be able to
            # access other scopes.
            if parent.scope == self.debusine.scope:
                yield parent.workspace

    def get_base_workflow_template(self) -> WorkflowTemplateData:
        """Find data for the base workflow template."""
        for workspace in self.iter_workspace_names():
            with self._api_call_or_fail():
                try:
                    base_workflow_template = (
                        self.debusine.workflow_template_get(
                            workspace, self.args.base_workflow_template
                        )
                    )
                    break
                except DebusineError:
                    pass
        else:
            self._fail(
                f"Workflow template {self.args.base_workflow_template!r} not "
                f"found in any parent workspace in this scope."
            )
        if base_workflow_template.task_name != "debian_pipeline":
            self._fail(
                f"Workflow template {base_workflow_template.name!r} is not a "
                f"debian_pipeline template."
            )
        return base_workflow_template

    def adjust_static_parameters(
        self, static_parameters: dict[str, Any]
    ) -> dict[str, Any]:
        """Adjust a workflow template's static parameters for the new suite."""
        static_parameters = dict(static_parameters)
        static_parameters.pop("enable_upload", None)
        static_parameters["suite"] = (
            f"{self.args.suite}@{CollectionCategory.SUITE}"
        )
        architectures = list(self.args.architectures)
        if "all" not in architectures:
            # This doesn't have to be present in the suite's data (if it's
            # missing, architecture-independent packages will be published
            # in the architecture-specific indexes), but omitting it from
            # the allowlist is likely to be confusing.
            architectures.insert(0, "all")
        static_parameters["architectures_allowlist"] = architectures
        return static_parameters

    @override
    def run(self) -> None:
        if self.args.base_workflow_template is not None:
            base_workflow_template = self.get_base_workflow_template()
        else:
            base_workflow_template = None

        collection_data = CollectionDataNew(
            name=self.args.suite,
            category=CollectionCategory.SUITE,
            data={
                "components": self.args.components or ["main"],
                "architectures": self.args.architectures,
            },
        )
        with self._api_call_or_fail():
            collection = self.debusine.collection_create(
                self.workspace, collection_data
            )
        output: dict[str, Any] = {
            "result": "success",
            "suite": self.debusine.webui_url(
                collection.get_link_url("webui_self")
            ),
        }

        if base_workflow_template is not None:
            workflow_template_data = WorkflowTemplateDataNew(
                name=f"publish-to-{self.args.suite}",
                task_name="debian_pipeline",
                static_parameters=self.adjust_static_parameters(
                    base_workflow_template.static_parameters
                ),
                runtime_parameters=base_workflow_template.runtime_parameters,
                priority=base_workflow_template.priority,
            )
            with self._api_call_or_fail():
                workflow_template = self.debusine.workflow_template_create(
                    self.workspace, workflow_template_data
                )
            output["workflow_template"] = self.debusine.webui_url(
                workflow_template.get_link_url("webui_self")
            )

        self.print_yaml_output(output)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""OpenSSL utilities for the Debusine signing service."""

import re
import subprocess
from pathlib import Path
from typing import BinaryIO, TYPE_CHECKING

from debusine.signing.signing_utils import SensitiveTemporaryDirectory

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer

    ReadableBuffer  # fake usage for vulture


def openssl_generate(
    common_name: str, certificate_days: int, log_file: BinaryIO | None = None
) -> tuple[bytes, bytes]:
    """Generate a new private key and certificate using OpenSSL."""
    with SensitiveTemporaryDirectory("debusine-openssl-generate-") as tmp:
        key = Path(tmp, "tmp.key")
        certificate = Path(tmp, "tmp.crt")
        # openssl-req(1ssl) says: 'Special characters may be escaped by "\"
        # (backslash), whitespace is retained.'  "/" and "=" are special
        # since they are used to delimit successive types and values.
        quoted_common_name = re.sub(r'([/=])', r'\\\1', common_name)
        subprocess.run(
            [
                "openssl",
                "req",
                "-new",
                "-newkey",
                "rsa:2048",
                "-x509",
                "-subj",
                f"/CN={quoted_common_name}/",
                "-days",
                str(certificate_days),
                "-noenc",
                "-sha256",
                "-keyout",
                key,
                "-out",
                certificate,
            ],
            check=True,
            stdout=log_file,
            stderr=log_file,
        )
        return key.read_bytes(), certificate.read_bytes()


def x509_fingerprint(
    certificate: "ReadableBuffer", log_file: BinaryIO | None = None
) -> str:
    """Get the SHA-256 fingerprint of an X.509 certificate."""
    return (
        subprocess.run(
            [
                "openssl",
                "x509",
                "-inform",
                "PEM",
                "-noout",
                "-fingerprint",
                "-sha256",
            ],
            check=True,
            input=certificate,
            stdout=subprocess.PIPE,
            stderr=log_file,
        )
        .stdout.decode()
        .rstrip("\n")
        # "sha256 Fingerprint=AA:BB:CC:..."
        .split("=", 1)[1]
        .replace(":", "")
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Transparently add scopes to URLs."""

import base64
import binascii
import datetime as dt
import re
from typing import TYPE_CHECKING

import django.http
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.utils import timezone
from django.utils.deprecation import MiddlewareMixin
from rest_framework.authentication import get_authorization_header

from debusine.db.permissioncontext import PermissionContext
from debusine.server.scopes import get_scope_urlconf

if TYPE_CHECKING:
    from debusine.db.models import Scope, Token

scope_prefix_re = re.compile(r'^/([^/]+)(/|$)')


class ScopeMiddleware(MiddlewareMixin):
    """
    Extract the current scope from the URL prefix.

    If used, it must be sequenced before
    django.middleware.common.CommonMiddleware, since it can make use of URL
    resolution.
    """

    def process_request(self, request: django.http.HttpRequest) -> None:
        """Process request in middleware."""
        from debusine.db.models.scopes import is_valid_scope_name

        scope_name: str | None = None

        if mo := scope_prefix_re.match(request.path_info):
            if mo.group(1) == "api":
                # /api/ gets special treatment, as scope can also be specified
                # in a header
                scope_name = request.headers.get("x-debusine-scope")
            elif is_valid_scope_name(mo.group(1)):
                scope_name = mo.group(1)

        if scope_name is None:
            scope_name = settings.DEBUSINE_DEFAULT_SCOPE

        request.debusine_scope = self.get_scope(scope_name)  # type: ignore[attr-defined]
        # VirtualHostMiddleware may have set this already; if so, leave it
        # alone, as the urlconfs it sets don't need to be per-scope.
        if getattr(request, "urlconf", None) is None:
            setattr(request, "urlconf", get_scope_urlconf(scope_name))

    def get_scope(self, name: str) -> "Scope":
        """Set the current scope to the given named one."""
        from django.shortcuts import get_object_or_404

        from debusine.db.models import Scope

        return get_object_or_404(Scope, name=name)


TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL = dt.timedelta(minutes=5)


class AuthorizationMiddleware(MiddlewareMixin):
    """
    Check user access to the current scope.

    If used, it must be sequenced after
    django.contrib.auth.middleware.AuthenticationMiddleware, since it needs the
    current user, and after
    debusine.server.middlewares.token_last_seen_at.TokenLastSeenAtMiddleware
    to validate the access of the worker token.
    """

    def _basic_auth_allowed(self, request: django.http.HttpRequest) -> bool:
        """Check if we should attempt HTTP basic authentication."""
        return request.scheme == "https"

    def _get_token_from_basic_auth(
        self, request: django.http.HttpRequest
    ) -> "Token | None":
        """
        Get a Debusine token sent as basic auth credentials.

        This is used to allow apt to pass Debusine tokens for repositories in
        private workspaces for archive views.
        """
        from debusine.db.models import Token

        if not getattr(request, "is_archive_view", False):
            return None

        # Force a Django request to d-r-f's get_authorization_header, which we
        # can do as it only really checks request.META
        auth = get_authorization_header(
            request  # type: ignore[arg-type]
        ).split(None, 1)
        if not auth or auth[0].lower() != b'basic':
            return None

        if not self._basic_auth_allowed(request):
            raise PermissionDenied(
                "Authenticated access is only allowed using HTTPS"
            )

        if len(auth) == 1:
            raise PermissionDenied("No credentials provided")

        try:
            try:
                auth_decoded = base64.b64decode(auth[1]).decode('utf-8')
            except UnicodeDecodeError:
                auth_decoded = base64.b64decode(auth[1]).decode('latin-1')

            userid, password = auth_decoded.split(':', 1)
        except (TypeError, ValueError, UnicodeDecodeError, binascii.Error):
            raise PermissionDenied("Credentials string is malformed")

        # Ensure that a token, and only the right kind of token, is passed via
        # basic auth
        token = Token.objects.get_token_or_none(password)
        if token is not None:
            token_username = (
                token.user.username
                if token.user is not None
                else token.external_username
            )
            if token_username == userid:
                return token
        # Intentionally give the same error message for all cases to avoid
        # leaking information on failed attempts.
        raise PermissionDenied("Invalid token")

    def _get_token_from_header(
        self, request: django.http.HttpRequest
    ) -> "Token | None":
        """Get a Debusine token from a token header."""
        from debusine.db.models import Token

        if (token_key := request.headers.get("token")) is None:
            return None

        return Token.objects.get_token_or_none(token_key)

    def _validate_and_cache_token(
        self, request: django.http.HttpRequest
    ) -> None:
        """Validate and cache the token in the request."""
        from debusine.db.models import Token

        setattr(request, "_debusine_token", None)

        token: Token | None = self._get_token_from_header(request)
        if token is None:
            token = self._get_token_from_basic_auth(request)
        if token is None:
            return

        now = timezone.now()
        if (
            token.last_seen_at is None
            or token.last_seen_at < now - TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL
        ):
            token.last_seen_at = now
            token.save(update_fields=["last_seen_at"])

        if not token.enabled:
            return

        setattr(request, "_debusine_token", token)

    def _get_permission_context(
        self, request: django.http.HttpRequest
    ) -> PermissionContext:
        """
        Create the permission context for a request.

        :returns: the permission context
        :raises PermissionDenied: in case of inconsistencies
        """
        from debusine.db.models import Token, WorkRequest

        if (token := getattr(request, "_debusine_token", None)) is None:
            if request.user.is_authenticated:
                return PermissionContext.for_ui_user(request.user)
            else:
                return PermissionContext.for_ui_anonymous()

        match token.token_type:
            case Token.Types.WORKER_ACTIVATION:
                return PermissionContext.for_worker_activation(token)
            case Token.Types.WORKER:
                return PermissionContext.for_worker(token)
            case Token.Types.WORK_REQUEST:
                assert token.work_request is not None
                if token.work_request.status != WorkRequest.Statuses.RUNNING:
                    raise PermissionDenied("work request is not running")
                return PermissionContext.for_work_request(token=token)
            case Token.Types.USER:
                if request.user.is_authenticated:
                    raise PermissionDenied(
                        "cannot use both Django and user token authentication"
                    )

                if not token.user.is_active:
                    raise PermissionDenied("user token has an inactive user")

                return PermissionContext.for_api_user(token)
            case Token.Types.EXTERNAL_USER:
                if request.user.is_authenticated:
                    raise PermissionDenied(
                        "cannot use both Django and user token authentication"
                    )

                return PermissionContext.for_api_external_user(token)
            case _ as unreachable:
                # Enforced by database constraint.
                raise AssertionError(f"Unexpected token type: {unreachable}")

    def process_request(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponseForbidden | None:
        """Process request in middleware."""
        from debusine.db.context import ContextConsistencyError, context
        from debusine.db.models import Scope

        if not hasattr(request, "debusine_scope"):
            raise ImproperlyConfigured(
                "AuthorizationMiddleware needs to be called"
                " after ScopeMiddleware"
            )
        assert isinstance(request.debusine_scope, Scope)

        try:
            self._validate_and_cache_token(request)
        except PermissionDenied as e:
            return django.http.HttpResponseForbidden(str(e))

        # request.user may be a lazy object, e.g. as set up by
        # AuthenticationMiddleware.  In that case we must force it to be
        # evaluated here, as otherwise asgiref.sync.AsyncToSync may try to
        # evaluate it when restoring context and will raise a
        # SynchronousOnlyOperation exception.
        request.user.username

        try:
            pc = self._get_permission_context(request)
        except PermissionDenied as e:
            return django.http.HttpResponseForbidden(str(e))

        try:
            context.set_permission_context_scoped(request.debusine_scope, pc)
        except ContextConsistencyError as e:
            return django.http.HttpResponseForbidden(str(e))

        return None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for TaskRunner."""

import asyncio
import datetime as dt
import logging
import queue
import stat
import time as tm
from collections.abc import Awaitable, Collection
from pathlib import Path
from typing import Literal, override
from unittest import mock

import debusine.tasks.models as task_models
from debusine.artifacts.models import (
    RuntimeStatistics,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.models import WorkRequestResponse
from debusine.tasks import BaseExternalTask, BaseTaskWithExecutor
from debusine.tasks.executors import (
    ExecutorInterface,
    ExecutorStatistics,
    InstanceInterface,
)
from debusine.tasks.models import WorkerType
from debusine.tasks.tests.helpers import SampleBaseExternalTask
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_work_request_response,
    preserve_task_registry,
)
from debusine.worker.config import ConfigHandler
from debusine.worker.system_information import cpu_time
from debusine.worker.task_runner import (
    CollectStatistics,
    TaskRunner,
    TaskRunnerError,
)
from debusine.worker.tests import server


class WorkerConfigTestCase(TestCase, server.DebusineAioHTTPTestCase):
    """Test case with a worker configuration."""

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.api_url = str(self.server.make_url('')) + "/api"

        self.temp_config_directory = self.create_temp_config_directory(
            {
                'General': {'api-url': self.api_url},
                'Environment': {'ENABLE': 'yes'},
            },
        )

        self.config = ConfigHandler(
            directories=[self.temp_config_directory.as_posix()]
        )
        self.config['General']['log-file'] = str(
            self.temp_config_directory / 'worker.log'
        )


class CollectStatisticsTests(TestCase, server.DebusineAioHTTPTestCase):
    """Tests for :py:class:`CollectStatistics`."""

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.statistics = RuntimeStatistics()
        self.task = mock.MagicMock(spec=BaseTaskWithExecutor)
        self.task.executor = None

    def set_executor(self, statistics: ExecutorStatistics) -> None:
        self.task.executor = mock.MagicMock(spec=ExecutorInterface)
        self.task.executor.get_statistics.return_value = statistics
        self.task.executor_instance = mock.MagicMock(spec=InstanceInterface)

    def test_init(self) -> None:
        collect = CollectStatistics(self.task, self.statistics)
        self.assertIs(collect.task, self.task)
        self.assertIs(collect.statistics, self.statistics)
        self.assertLessEqual(collect.start_time, tm.monotonic())
        self.assertLessEqual(collect.start_cpu_time, cpu_time())
        self.assertFalse(collect.done.is_set())
        self.assertIsNone(collect.collect_task)

    def test_gather_no_executor(self) -> None:
        collect = CollectStatistics(self.task, self.statistics)
        collect.gather()
        self.assertEqual(self.statistics, RuntimeStatistics())

    def test_gather(self) -> None:
        es = ExecutorStatistics(
            disk_space=1,
            memory=2,
            available_disk_space=4,
            available_memory=8,
            cpu_count=16,
        )
        self.set_executor(es)
        collect = CollectStatistics(self.task, self.statistics)
        with self.spy(
            "debusine.tasks.executors.base.ExecutorStatistics.add_to"
        ) as add_to:
            collect.gather()
        self.assertEqual(
            self.statistics,
            RuntimeStatistics(
                disk_space=1,
                memory=2,
                available_disk_space=4,
                available_memory=8,
                cpu_count=16,
            ),
        )
        add_to.assert_called_once_with(es, self.statistics)

    async def test_collect(self) -> None:
        collect = CollectStatistics(self.task, self.statistics)
        self.assertIsNone(self.statistics.duration)
        self.assertIsNone(self.statistics.cpu_time)

        gather = self.enterContext(mock.patch.object(collect, "gather"))

        count = 0

        async def mock_wait(
            _aws: Collection[Awaitable[None]],
            timeout: int,  # noqa: ASYNC109, ARG001
        ) -> None:
            nonlocal count
            count += 1
            if count == 3:
                collect.done.set()

        self.enterContext(mock.patch("asyncio.wait", side_effect=mock_wait))

        await collect.collect()

        self.assertEqual(gather.call_count, 3)

        self.assertIsNotNone(self.statistics.duration)
        self.assertIsNotNone(self.statistics.cpu_time)

    async def test_context_manager(self) -> None:
        collect = CollectStatistics(self.task, self.statistics)
        gather = self.enterContext(mock.patch.object(collect, "gather"))

        self.assertIsNone(collect.collect_task)
        self.assertFalse(collect.done.is_set())

        async with collect:
            self.assertIsNotNone(collect.collect_task)

        self.assertGreaterEqual(gather.call_count, 1)

        self.assertTrue(collect.done.is_set())
        self.assertIsNone(collect.collect_task)


class TaskRunnerTests(WorkerConfigTestCase):
    """Tests for :py:class:`TaskRunner`."""

    runner: TaskRunner

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.workdir = self.config.working_directory
        self.workdir.mkdir(parents=True, exist_ok=True)
        self.worker_type: Literal[WorkerType.EXTERNAL, WorkerType.SIGNING] = (
            WorkerType.EXTERNAL
        )
        self.work_request = WorkRequestResponse(
            id=42,
            url=self.api_url + "/work_requests/42",
            created_at=dt.datetime(2026, 1, 1, 0, tzinfo=dt.UTC),
            started_at=dt.datetime(2026, 1, 1, 1, tzinfo=dt.UTC),
            status="running",
            result="",
            worker=1,
            task_type="Worker",
            task_name="noop",
            task_data={},
            token="work-request-token",
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            artifacts=[],
            scope="debusine",
            workspace="Playground",
        )

    @override
    async def tearDownAsync(self) -> None:
        await self.runner.shutdown()
        await super().tearDownAsync()

    def setup_runner(self) -> None:
        """Set up a runner for this work request."""
        self.runner = TaskRunner(
            self.config, self.worker_type, self.work_request, self.workdir
        )

    def test_init(self) -> None:
        self.setup_runner()
        self.assertIs(self.runner.config, self.config)
        self.assertIs(self.runner.worker_type, self.worker_type)
        self.assertIs(self.runner.work_request, self.work_request)
        self.assertTrue(self.runner.workdir.is_relative_to(self.workdir))
        st = self.runner.workdir.stat()
        self.assertEqual(stat.S_IMODE(st.st_mode), 0o755)
        self.assertEqual(self.runner.client.base_api_url, self.config.api_url)
        self.assertEqual(
            self.runner.client.token, self.runner.work_request.token
        )
        self.assertEqual(self.runner.client.timeout, self.config.http_timeout)
        self.assertEqual(self.runner.client.scope, self.work_request.scope)
        self.assertIsNone(self.runner.task)

    async def test_setup_task(self) -> None:
        set_working_directory = self.enterContext(
            mock.patch.object(BaseExternalTask, "set_working_directory")
        )
        set_worker_environment_variables = self.enterContext(
            mock.patch.object(
                BaseExternalTask, "set_worker_environment_variables"
            )
        )
        configure_server_access = self.enterContext(
            mock.patch.object(BaseExternalTask, "configure_server_access")
        )

        self.setup_runner()
        await self.runner.setup_task()
        task = self.runner.task
        assert task is not None

        set_working_directory.assert_called_once_with(self.runner.workdir)
        set_worker_environment_variables.assert_called_once_with(
            {"ENABLE": "yes"}
        )
        configure_server_access.assert_called_once_with(self.runner.client)
        self.assertEqual(task.work_request_id, self.work_request.id)
        self.assertEqual(task.workspace_name, self.work_request.workspace)

    async def test_setup_task_signing_task_on_external_worker(self) -> None:
        self.enterContext(preserve_task_registry())

        class SigningTestData(task_models.BaseTaskData):
            pass

        class SigningTest(
            SampleBaseExternalTask[
                SigningTestData, task_models.BaseDynamicTaskData
            ],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SIGNING

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
                raise NotImplementedError()

        self.work_request = create_work_request_response(
            id=52,
            task_type="Signing",
            task_name="signingtest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            token="work-request-token",
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.setup_runner()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.setup_task()

        self.assertEqual(exc.exception.code, "wrong-task-type")
        self.assertEqual(
            exc.exception.message, "Task: %s is of type %s, not %s"
        )
        self.assertEqual(
            exc.exception.args,
            (
                self.work_request.task_name,
                self.work_request.task_type,
                "Worker",
            ),
        )
        self.assertIsNone(self.runner.task)

    async def test_setup_task_worker_task_on_signing_worker(self) -> None:
        self.worker_type = WorkerType.SIGNING
        self.setup_runner()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.setup_task()

        self.assertEqual(exc.exception.code, "wrong-task-type")
        self.assertEqual(
            exc.exception.message, "Task: %s is of type %s, not %s"
        )
        self.assertEqual(
            exc.exception.args,
            (
                self.work_request.task_name,
                self.work_request.task_type,
                "Signing",
            ),
        )
        self.assertIsNone(self.runner.task)

    async def test_setup_task_signing_task_on_signing_worker(self) -> None:
        self.enterContext(preserve_task_registry())

        class SigningTestData(task_models.BaseTaskData):
            pass

        class SigningTest(
            SampleBaseExternalTask[
                SigningTestData, task_models.BaseDynamicTaskData
            ],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SIGNING

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
                raise NotImplementedError()

        self.worker_type = WorkerType.SIGNING
        self.work_request = create_work_request_response(
            id=52,
            task_type="Signing",
            task_name="signingtest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            token="work-request-token",
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.setup_runner()
        await self.runner.setup_task()
        self.assertIsNotNone(self.runner.task)

    async def test_setup_task_server_task_on_external_worker(self) -> None:
        self.enterContext(preserve_task_registry())

        class ServerTestData(task_models.BaseTaskData):
            pass

        class ServerTest(
            SampleBaseExternalTask[
                ServerTestData, task_models.BaseDynamicTaskData
            ],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SERVER

            @override
            def _execute(self) -> WorkRequestResults:
                raise NotImplementedError()

            @override
            def _upload_work_request_debug_logs(self) -> None:
                raise NotImplementedError()

        self.work_request = create_work_request_response(
            id=52,
            task_type="Server",
            task_name="servertest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            token="work-request-token",
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )

        self.setup_runner()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.setup_task()

        self.assertEqual(exc.exception.code, "wrong-task-type")
        self.assertEqual(
            exc.exception.message, "Task: %s is of type %s, not %s"
        )
        self.assertEqual(
            exc.exception.args,
            (
                self.work_request.task_name,
                self.work_request.task_type,
                "Worker",
            ),
        )
        self.assertIsNone(self.runner.task)

    async def test_setup_task_unknown_task_name(self) -> None:
        self.work_request = create_work_request_response(
            id=52,
            task_type="Worker",
            task_name="nonexistent",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            token="work-request-token",
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.setup_runner()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.setup_task()

        self.assertEqual(exc.exception.code, "setup-failed")
        self.assertEqual(exc.exception.message, "Task: %s Error setup: %s")
        self.assertEqual(exc.exception.args[0], self.work_request.task_name)
        self.assertEqual(
            str(exc.exception.args[1]),
            "'nonexistent' is not a registered Worker task_name",
        )
        self.assertIsNone(self.runner.task)

    async def test_setup_task_invalid_task_data(self) -> None:
        self.work_request.task_data = {"invalid": True}
        self.setup_runner()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.setup_task()

        self.assertEqual(exc.exception.code, "configure-failed")
        self.assertEqual(exc.exception.message, "Task: %s Error configure: %s")
        self.assertEqual(exc.exception.args[0], self.work_request.task_name)
        self.assertRegex(
            str(exc.exception.args[1]), r"Extra inputs are not permitted"
        )
        self.assertIsNone(self.runner.task)

    async def test_execute(self) -> None:
        gather = self.enterContext(
            mock.patch.object(CollectStatistics, "gather")
        )
        execute = self.enterContext(
            mock.patch(
                "debusine.tasks.Noop.execute_logging_exceptions",
                return_value=WorkRequestResults.SUCCESS,
            )
        )
        self.setup_runner()
        await self.runner.setup_task()
        self.assertEqual(
            await self.runner.execute(), WorkRequestResults.SUCCESS
        )

        gather.assert_called()
        execute.assert_called()

    async def test_execute_raises(self) -> None:
        gather = self.enterContext(
            mock.patch.object(CollectStatistics, "gather")
        )
        self.enterContext(
            mock.patch(
                "debusine.tasks.Noop.execute_logging_exceptions",
                side_effect=RuntimeError("expected error"),
            )
        )
        self.setup_runner()
        await self.runner.setup_task()
        with self.assertRaises(TaskRunnerError) as exc:
            await self.runner.execute()

        gather.assert_called()

        self.assertEqual(exc.exception.code, "execute-failed")
        self.assertEqual(exc.exception.message, "Task: %s Error execute: %s")
        self.assertEqual(exc.exception.args[0], self.work_request.task_name)
        self.assertEqual(str(exc.exception.args[1]), "expected error")

    async def test_cancel_execution_aborts_task(self) -> None:
        self.enterContext(mock.patch.object(CollectStatistics, "gather"))

        loop = asyncio.get_running_loop()
        executing = asyncio.Event()
        test_done: queue.Queue[None] = queue.Queue()

        def mock_task_execute() -> WorkRequestResults:
            nonlocal executing
            loop.call_soon_threadsafe(executing.set)
            test_done.get()
            return WorkRequestResults.SUCCESS

        self.enterContext(
            mock.patch(
                "debusine.tasks.noop.Noop.execute_logging_exceptions",
                side_effect=mock_task_execute,
            )
        )
        self.setup_runner()
        await self.runner.setup_task()

        # Wait until the runner starts executing the task
        execution = asyncio.create_task(self.runner.execute())
        await executing.wait()
        assert self.runner.task is not None
        self.assertFalse(self.runner.task.aborted)

        # Cancel the task execution
        execution.cancel()
        await asyncio.wait([execution])

        self.assertTrue(self.runner.task.aborted)
        test_done.put_nowait(None)

    async def test_context_manager(self) -> None:
        gather = self.enterContext(
            mock.patch.object(CollectStatistics, "gather")
        )
        execute = self.enterContext(
            mock.patch(
                "debusine.tasks.Noop.execute_logging_exceptions",
                return_value=WorkRequestResults.SUCCESS,
            )
        )
        self.setup_runner()

        self.assertIsNone(self.runner.task)
        async with self.runner:
            await self.runner.setup_task()
            self.assertIsNotNone(self.runner.task)
            self.assertTrue(self.runner.workdir.is_dir())
            self.assertEqual(
                await self.runner.execute(), WorkRequestResults.SUCCESS
            )
        self.assertFalse(self.runner.workdir.is_dir())

        gather.assert_called()
        execute.assert_called()

    async def test_shutdown_cleanup_error(self) -> None:
        self.setup_runner()

        with self.assertLogs(
            logger="debusine.worker.task_runner", level=logging.WARNING
        ) as log:
            async with self.runner:
                with mock.patch.object(
                    self.runner.workdir_tempdir,
                    "cleanup",
                    side_effect=RuntimeError("expected error"),
                ):
                    await self.runner.shutdown()

        [warning] = log.output
        self.assertRegex(
            warning,
            r"WARNING:debusine.worker.task_runner:"
            "Failed to remove '/.+/WR-42-.+':"
            r" expected error",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for debusine.tests.base.TestCase."""

import contextlib
import datetime as dt
import io
import logging
import re
from configparser import ConfigParser
from pathlib import Path
from unittest import mock

import lxml
import requests
import responses

from debusine.test.base import TestCase


class TestCaseTests(TestCase):
    """Tests for methods in debusine.test.base.TestCase."""

    def test_create_temp_config_directory(self) -> None:
        """create_temp_config_directory write the configuration."""
        config = {
            'General': {'default-server': 'debian'},
            'server:debian': {
                'url': 'https://debusine.debian.org',
                'token': 'token-for-debian',
            },
        }
        directory = self.create_temp_config_directory(config)

        actual_config = ConfigParser()
        actual_config.read(directory / 'config.ini')

        expected_config = ConfigParser()
        expected_config.read_dict(config)

        self.assertEqual(actual_config, expected_config)

    def test_assert_dict_contains_subset_raises_exception(self) -> None:
        """Raise an exception (subset not in dictionary)."""
        expected_message = "{'b': 1} does not contain the subset {'a': 1}"
        with self.assertRaisesRegex(self.failureException, expected_message):
            self.assertDictContainsAll({'b': 1}, {'a': 1})

    def test_assert_dict_use_error_msg(self) -> None:
        """Raise exception using a specific error message."""
        expected_message = 'Missing values'

        with self.assertRaisesRegex(self.failureException, expected_message):
            self.assertDictContainsAll({}, {'a': 1}, expected_message)

    def test_assert_dict_contains_subset_does_not_raise_exception(self) -> None:
        """Do not raise any exception (subset in dictionary)."""
        self.assertDictContainsAll({'a': 1, 'b': 2}, {'a': 1})

    def test_assert_dict_contains_subset_arg1_not_a_dictionary(self) -> None:
        """Raise exception because of wrong type argument 1."""
        expected_message = (
            "'a' is not an instance of <class 'collections.abc.Mapping'> : "
            "First argument is not a mapping"
        )
        with self.assertRaisesRegex(
            self.failureException, re.escape(expected_message)
        ):
            self.assertDictContainsAll('a', {})  # type: ignore[arg-type]

    def test_assert_dict_contains_subset_arg2_not_a_dictionary(self) -> None:
        """Raise exception because of wrong type argument 2."""
        expected_message = (
            "'b' is not an instance of <class 'collections.abc.Mapping'> : "
            "Second argument is not a mapping"
        )
        with self.assertRaisesRegex(
            self.failureException, re.escape(expected_message)
        ):
            self.assertDictContainsAll({}, 'b')  # type: ignore[arg-type]

    def test_assert_raises_system_exit_no_system_exit(self) -> None:
        """Raise self.failureException because missing SystemExit."""
        expected_message = (
            r'SystemExit not raised : Did not raise '
            r'SystemExit with exit_code=\^3\$'
        )
        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertRaisesSystemExit(3),
        ):
            pass

    def test_assert_raises_system_exit_unexpected_exit_code(self) -> None:
        """Raise self.failureException because wrong exit_code in SystemExit."""
        expected_message = (
            r'\^3\$" does not match "7" : Did not raise '
            r'SystemExit with exit_code=\^3\$'
        )
        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertRaisesSystemExit(3),
        ):
            raise SystemExit(7)

    def test_assert_raises_system_exit_success(self) -> None:
        """Do not raise self.failureException: expected SystemExit is raised."""
        with self.assertRaisesSystemExit(3):
            raise SystemExit(3)

    @responses.activate
    def test_assert_token_key_included_in_all_requests(self) -> None:
        """Do not raise any exception (all requests had the token key)."""
        responses.add(
            responses.GET,
            'https://example.net/something',
        )
        responses.add(
            responses.GET,
            'https://example.net/something',
        )

        token_key = 'some-key'

        requests.get(
            'https://example.net/something',
            headers={'Token': token_key},
        )

        self.assert_token_key_included_in_all_requests(token_key)

    @responses.activate
    def test_assert_token_key_included_in_all_requests_raise_missing(
        self,
    ) -> None:
        """Raise exception because token not included in the request."""
        responses.add(
            responses.GET,
            'https://example.net/something',
        )

        requests.get('https://example.net/something')

        expected_message = (
            "Token missing in the headers for the request "
            "'https://example.net/something'"
        )

        with self.assertRaisesRegex(self.failureException, expected_message):
            self.assert_token_key_included_in_all_requests('some-token')

    @responses.activate
    def test_assert_token_key_included_in_all_requests_raise_mismatch(
        self,
    ) -> None:
        """Raise exception because token mismatch included in the request."""
        responses.add(
            responses.GET,
            'https://example.net/something',
        )

        token = 'token-for-server'

        requests.get(
            'https://example.net/something',
            headers={'Token': 'some-invalid-token'},
        )

        expected_message = (
            "Unexpected token. In the request: "
            "'https://example.net/something' "
            "Actual: 'some-invalid-token' Expected: 'token-for-server'"
        )

        with self.assertRaisesRegex(self.failureException, expected_message):
            self.assert_token_key_included_in_all_requests(token)

    def test_assertLogsContains_log_found(self) -> None:
        """assertLogsContains() does not raise self.failureException."""
        with self.assertLogsContains('required-log') as logs:
            logging.warning('required-log')

        self.assertEqual(logs.output, ["WARNING:root:required-log"])

    def test_assertLogsContains_log_expected_count_wrong(self) -> None:
        """assertLogsContains() raise self.failureException (wrong count)."""
        expected_message = (
            '^Expected: "required-log"\n'
            'Actual: "WARNING:root:required-log"\n'
            'Expected msg found 1 times, expected 2 times$'
        )

        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertLogsContains('required-log', expected_count=2),
        ):
            logging.warning('required-log')

    def test_assertLogsContains_log_expected_not_found(self) -> None:
        """assertLogsContains() raise self.failureException (wrong count)."""
        expected_message = (
            '^Expected: "required-log"\n'
            'Actual: "WARNING:root:some-log"\n'
            'Expected msg found 0 times, expected 1 times$'
        )

        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertLogsContains('required-log'),
        ):
            logging.warning('some-log')

    def test_assertLogsContains_log_expected_not_found_wrong_level(
        self,
    ) -> None:
        """assertLogsContains() raise self.failureException (wrong level)."""
        expected_message = (
            'no logs of level WARNING or higher triggered on root'
        )

        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertLogsContains('required-log', level=logging.WARNING),
        ):
            logging.debug('some-log')

    def test_assertLogsContains_log_not_found_in_raise_exception(self) -> None:
        """
        assertLogsContains() raise self.failureException.

        It handles raised exceptions in the context code.
        """
        expected_message = (
            '^Expected: "The wanted message"\n'
            'Actual: "WARNING:root:Unrelated message"\n'
            'Expected msg found 0 times, expected 1 times$'
        )

        with (
            self.assertRaisesRegex(self.failureException, expected_message),
            self.assertRaisesRegex(SystemExit, '3'),
            self.assertLogsContains('The wanted message'),
        ):
            logging.warning('Unrelated message')
            raise SystemExit(3)

    def test_assertLogsContains_forwards_exception(self) -> None:
        msg = "The wanted message"
        with (
            self.assertRaises(RuntimeError),
            self.assertLogsContains(msg),
        ):
            logging.warning(msg)
            raise RuntimeError("expected error")

    def test_dump_html(self) -> None:
        """Test dumping HTML to stdout."""
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            self.dump_html("<test>")
        self.assertEqual(stdout.getvalue().splitlines(), ["001 <test>"])

    def test_dump_html_with_title(self) -> None:
        """Test dumping HTML to stdout, with title."""
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            self.dump_html("<test>", title="title")
        self.assertEqual(
            stdout.getvalue().splitlines(), ["", "    title", "", "001 <test>"]
        )

    def test_dump_element(self) -> None:
        """Test dumping an element to stdout."""
        content = b"<!DOCTYPE html><html></html>"
        tree = self.assertHTMLValid(content)

        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            self.dump_element(tree, title="title")
            stdout.getvalue().splitlines(), ["", "    title", "", "001 <html/>"]

    def test_dump_element_with_title(self) -> None:
        """Test dumping HTML to stdout, with title."""
        content = b"<!DOCTYPE html><html></html>"
        tree = self.assertHTMLValid(content)

        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            self.dump_element(tree)
        self.assertEqual(stdout.getvalue().splitlines(), ["001 <html/>"])

    def test_invalid_html(self) -> None:
        """Test warnings on invalid HTML."""
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            content = b"<!DOCTYPE html><html></nothtml>"
            with self.assertRaisesRegex(
                AssertionError,
                r"1:ERR_TAG_NAME_MISMATCH:Unexpected end tag : nothtml",
            ):
                self.assertHTMLValid(content)

        self.assertEqual(stdout.getvalue(), "")

    def test_invalid_html_dump_on_error(self) -> None:
        """Test warnings on invalid HTML with dump_on_error."""
        content = b"<!DOCTYPE html><html>\n</nothtml>"

        stdout = io.StringIO()
        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesRegex(
                AssertionError,
                r"2:ERR_TAG_NAME_MISMATCH:Unexpected end tag : nothtml",
            ),
        ):
            self.assertHTMLValid(content, dump_on_error=True)

        lines = stdout.getvalue().splitlines()
        self.assertEqual(lines[1], "    HTML parser errors")
        self.assertEqual(lines[3], "001 <!DOCTYPE html><html>")
        self.assertEqual(lines[4], "002 </nothtml>")
        self.assertEqual(
            lines[5], "  → ERR_TAG_NAME_MISMATCH:Unexpected end tag : nothtml"
        )

    def test_assert_has_element(self) -> None:
        """Test that assertHasElement works as expected."""
        tree = lxml.objectify.fromstring("<p><b>this</b> is a <b>test</b></p>")

        p = self.assertHasElement(tree, "//p")
        self.assertTextContentEqual(p, "this is a test")

        with self.assertRaisesRegex(AssertionError, "'//i' not found in tree"):
            self.assertHasElement(tree, "//i")
        with self.assertRaisesRegex(
            AssertionError, "'//b' matched 2 elements instead of one"
        ):
            self.assertHasElement(tree, "//b")

    def test_assert_has_element_dump_on_error(self) -> None:
        """Test assertHasElement with dump_on_error."""
        tree = lxml.objectify.fromstring("<p><b>this</b> is a <b>test</b></p>")

        for xpath, message in (
            ("//i", "'//i' not found in tree"),
            ("//b", "'//b' matched 2 elements instead of one"),
        ):
            with self.subTest(xpath=xpath):
                with (
                    mock.patch(
                        "debusine.test.base.TestCase.dump_element"
                    ) as dump,
                    self.assertRaisesRegex(AssertionError, re.escape(message)),
                ):
                    self.assertHasElement(tree, xpath, dump_on_error=True)
                dump.assert_called_with(tree, title=message)

    def test_assertElementHasClass(self) -> None:
        """Test assertElementHasClass."""
        for html, message in (
            ("<a></a>", "Element does not have a class attribute"),
            ("<a class=''></a>", "'foo' not found in []"),
            ("<a class='bar'></a>", "'foo' not found in ['bar']"),
            ("<a class='foo'></a>", None),
            ("<a class='foo bar'></a>", None),
            ("<a class='bar foo'></a>", None),
        ):
            with self.subTest(html=html):
                tree = self.assertHTMLValid(html)
                el = self.assertHasElement(tree, "body/a")
                if message is None:
                    self.assertElementHasClass(el, "foo")
                else:
                    with self.assertRaisesRegex(
                        AssertionError, re.escape(message)
                    ):
                        self.assertElementHasClass(el, "foo")

    def test_assertElementHasNoClass(self) -> None:
        """Test assertElementHasClass."""
        for html, message in (
            ("<a></a>", None),
            ("<a class=''></a>", None),
            ("<a class='bar'></a>", None),
            ("<a class='foo'></a>", "'foo' unexpectedly found in ['foo']"),
            (
                "<a class='bar foo'></a>",
                "'foo' unexpectedly found in ['bar', 'foo']",
            ),
            (
                "<a class='bar foo bar'></a>",
                "'foo' unexpectedly found in ['bar', 'foo', 'bar']",
            ),
        ):
            with self.subTest(html=html):
                tree = self.assertHTMLValid(html)
                el = self.assertHasElement(tree, "body/a")
                if message is None:
                    self.assertElementHasNoClass(el, "foo")
                else:
                    with self.assertRaisesRegex(
                        AssertionError, re.escape(message)
                    ):
                        self.assertElementHasNoClass(el, "foo")

    def test_assertTextContentEqual_dump_on_error(self) -> None:
        """Test assertTextContentEqual with dump_on_error."""
        tree = lxml.objectify.fromstring("<p><b>this</b> is a <b>test</b></p>")

        with (
            mock.patch("debusine.test.base.TestCase.dump_element") as dump,
            self.assertRaisesRegex(
                AssertionError, r"'this is a test' != 'mismatch'"
            ),
        ):
            self.assertTextContentEqual(tree, "mismatch", dump_on_error=True)
        dump.assert_called_with(tree, title="Text content mismatch")

    def test_assertDatetimeContentEqual(self) -> None:
        """Test :py:meth:`assertDatetimeContentEqual`."""
        tree = self.assertHTMLValid(
            "<p>created on <b>2025-01-01 12:30</b> at lunchtime</p>"
        )
        self.assertDatetimeContentEqual(
            tree.xpath("//b")[0],
            dt.datetime(2025, 1, 1, 12, 30, tzinfo=dt.UTC),
        )

    def test_assertDatetimeContentEqual_anywhere(self) -> None:
        """Test :py:meth:`assertDatetimeContentEqual`."""
        tree = self.assertHTMLValid(
            "<p>created on <b>2025-01-01 12:30</b> at lunchtime</p>"
        )
        self.assertDatetimeContentEqual(
            tree.xpath("//p")[0],
            dt.datetime(2025, 1, 1, 12, 30, tzinfo=dt.UTC),
            anywhere=True,
        )

    def test_assertDatetimeContentEqualNoDatetime(self) -> None:
        """Test :py:meth:`assertDatetimeContentEqual` with no datetime."""
        tree = self.assertHTMLValid("<ul><li>none</li></ul>")
        with self.assertRaisesRegex(
            self.failureException,
            r"text content 'none' does not contain"
            r" a YYYY-MM-DD HH:MM datetime",
        ):
            self.assertDatetimeContentEqual(
                tree.xpath("//ul")[0],
                dt.datetime(2025, 1, 1, 12, 30, tzinfo=dt.UTC),
            )

    def test_assertDatetimeContentEqualDifferentDatetime(self) -> None:
        """Test :py:meth:`assertDatetimeContentEqual` with another datetime."""
        tree = self.assertHTMLValid("<ul><li>2025-01-01 12:31</li></ul>")
        with self.assertRaisesRegex(
            self.failureException,
            r"datetime.datetime\(2025, 1, 1, 12, 31, tzinfo=\S+\)"
            r" != datetime.datetime\(2025, 1, 1, 12, 30, tzinfo=\S+\)",
        ):
            self.assertDatetimeContentEqual(
                tree.xpath("//ul")[0],
                dt.datetime(2025, 1, 1, 12, 30, tzinfo=dt.UTC),
            )

    def test_assertDatetimeContentEqualIgnoreSeconds(self) -> None:
        """:py:meth:`assertDatetimeContentEqual` ignores seconds."""
        tree = self.assertHTMLValid("<ul><li>2025-01-01 12:30</li></ul>")
        for expected in (
            dt.datetime(2025, 1, 1, 12, 30, 0, tzinfo=dt.UTC),
            dt.datetime(2025, 1, 1, 12, 30, 1, tzinfo=dt.UTC),
            dt.datetime(2025, 1, 1, 12, 30, 1, 12345, tzinfo=dt.UTC),
            dt.datetime(2025, 1, 1, 12, 30, 0, 12345, tzinfo=dt.UTC),
        ):
            with self.subTest(expected=expected.isoformat()):
                self.assertDatetimeContentEqual(tree.xpath("//ul")[0], expected)

    def test_assertYAMLContentEqual(self) -> None:
        """Test :py:meth:`assertYAMLContentEqual`."""
        from debusine.web.views.view_utils import highlight_as_yaml

        for data in (
            None,
            42,
            "test",
            [],
            {},
            [1, 2, 3],
            ["one", "two", "three"],
            {1: "one", "two": 2},
        ):
            with self.subTest(data=data):
                html = highlight_as_yaml(data)
                tree = self.assertHTMLValid(html)
                self.assertYAMLContentEqual(tree, data)

    def test_assertYAMLContentEqualBadYAML(self) -> None:
        """Test :py:meth:`assertYAMLContentEqual` with bad YAML."""
        for text in (
            "[",
            "{",
        ):
            with self.subTest(text=text):
                tree = self.assertHTMLValid(f"<pre>{text}</pre>")
                with self.assertRaisesRegex(
                    self.failureException,
                    r"node text .+ does not parse as YAML",
                ):
                    self.assertYAMLContentEqual(tree, {})

    def test_spy(self) -> None:
        with self.spy(
            "debusine.test.base.TestCase.create_temporary_file"
        ) as meth:
            path = self.create_temporary_file(suffix=".test")
            self.assertIsInstance(path, Path)
            self.assertEqual(path.name[-5:], ".test")
            meth.assert_called_once_with(self, suffix=".test")

    def test_spy_invalid_name(self) -> None:
        target = "invalid_method_name"
        with (
            self.assertRaisesRegex(
                TypeError,
                r"Need a valid target to patch\. You supplied:"
                f" {re.escape(repr(target))}",
            ),
            self.spy(target),
        ):
            self.fail("This should never be reached")  # pragma: no cover

    def test_spy_object(self) -> None:
        with self.spy_object(self, "create_temporary_file") as meth:
            path = self.create_temporary_file(suffix=".test")
            self.assertIsInstance(path, Path)
            self.assertEqual(path.name[-5:], ".test")
            meth.assert_called_once_with(suffix=".test")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli Artifact commands."""

import datetime as dt
import math
import os
import textwrap
from itertools import count
from pathlib import Path
from typing import Any, Literal
from unittest import mock
from unittest.mock import MagicMock
from urllib.parse import quote, urljoin

import yaml
from debian.deb822 import Dsc

from debusine.artifacts import LocalArtifact
from debusine.artifacts.models import ArtifactCategory
from debusine.artifacts.playground import ArtifactPlayground
from debusine.client import exceptions
from debusine.client.commands.base import DebusineCommand
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    ArtifactResponse,
    RelationResponse,
    RelationType,
    RelationsResponse,
    RemoteArtifact,
)
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
)


class CliCreateArtifactTests(BaseCliTests):
    """Tests for the Cli functionality related to create artifacts."""

    def create_files_to_upload(self) -> list[Path]:
        """Create three files to upload in a temporary directory."""
        temp_directory = self.create_temporary_directory()
        debian_tar_path = temp_directory / "pkg_1.0-1.debian.tar.gz"
        debian_tar_path.write_bytes(b"test")
        orig_tar_path = temp_directory / "pkg_1.0-1.orig.tar.gz"
        orig_tar_path.write_bytes(b"test2")
        dsc_path = temp_directory / "pkg_1.0-1.dsc"
        ArtifactPlayground.write_deb822_file(
            Dsc,
            dsc_path,
            [debian_tar_path, orig_tar_path],
            source="pkg",
            version="1.0-1",
        )
        return [dsc_path, debian_tar_path, orig_tar_path]

    def patch_create_artifact(self) -> MagicMock:
        """Patch Debusine.create_artifact."""
        patcher_create_artifact = mock.patch.object(
            Debusine, "artifact_create", autospec=True
        )
        mocked_create_artifact = patcher_create_artifact.start()

        self.addCleanup(patcher_create_artifact.stop)
        return mocked_create_artifact

    def assert_artifact_is_created(
        self,
        mocked_create_artifact: MagicMock,
        expected_workspace: str,
        expected_artifact_id: int,
        expected_files_to_upload_count: int,
        stdout: str,
        stderr: str,
        expected_expire_at: dt.datetime | None = None,
        data_from_stdin: bool = False,
    ) -> LocalArtifact[Any]:
        """
        Assert that mocked_create_artifact was called only once and stdout.

        In stdout expects that CLI printed into stdout the information
        about the recently created artifact.

        Return LocalArtifact that would be sent to the server.

        :param mocked_create_artifact: mock to assert that was called only once.
        :param expected_workspace: expected created workspace.
        :param expected_artifact_id: expected created artifact id.
        :param expected_files_to_upload_count: expected files to upload.
        :param stdout: contains Cli.execute() output to verify expected output.
        :param stderr: contains Cli.execute() output to verify expected output.
        :param expected_expire_at: expected artifact expiry time.
        :param data_from_stdin: if True, input data came from stdin.
        """
        mocked_create_artifact.assert_called_once()

        base_url = self.get_base_url(self.default_server)
        scope = self.servers[self.default_server]["scope"]
        artifact_url = urljoin(
            base_url,
            f"{quote(scope)}/{quote(expected_workspace)}/"
            f"artifact/{expected_artifact_id}/",
        )

        expected = yaml.safe_dump(
            {
                "result": "success",
                "message": f"New artifact created: {artifact_url}",
                "artifact_id": expected_artifact_id,
                "files_to_upload": expected_files_to_upload_count,
                "expire_at": expected_expire_at,
            },
            sort_keys=False,
            width=math.inf,
        )
        if data_from_stdin:
            expected = f"---\n{expected}"

        self.assertEqual(stdout, expected)
        self.assertEqual(stderr, "")

        local_artifact = mocked_create_artifact.call_args[0][1]
        assert isinstance(local_artifact, LocalArtifact)

        return local_artifact

    def create_artifact_via_cli(
        self,
        category: str,
        *,
        workspace: str | None = None,
        files_to_upload: list[str] | None = None,
        upload_base_directory: Path | None = None,
        expire_at: dt.datetime | None = None,
        data: bytes | None = None,
        data_file: Path | Literal["-"] | None = None,
        stdin: dict[str, Any] | None = None,
        create_artifact_return_value: ArtifactResponse | None = None,
        create_artifact_side_effect: Any = None,
        assert_system_exit_code: int | None = None,
    ) -> tuple[str, str, MagicMock]:
        """
        Call CLI create-artifact with the correct parameters.

        Patch Debusine.create_artifact before executing the command.

        :param category: category of the artifact.
        :param workspace: add ["--workspace", workspace] options.
        :param files_to_upload: list of file paths to upload.
        :param expire_at: datetime (for --expire-at option).
        :param data: data passed to the artifact (creates a temporary file
          and passes the temporary file path via --data).
        :param data_file: data file (for --data options).
        :param stdin: dump stdin into YAML and writes it into stdin.
        :param create_artifact_return_value: return value for the
          Debusine.create_artifact mock.
        :param create_artifact_side_effect: side effect for the
          Debusine.create_artifact mock.
        :param assert_system_exit_code: assert that the create-artifact
          command raises SystemExit(X).

        Return tuple with stderr, stdout and the patched
        Debusine.create_artifact
        """
        mocked_create_artifact = self.patch_create_artifact()

        if create_artifact_return_value is not None:
            mocked_create_artifact.return_value = create_artifact_return_value

        if create_artifact_side_effect is not None:
            mocked_create_artifact.side_effect = create_artifact_side_effect

        create_cli_options = [category, "--yaml"]

        if workspace is not None:
            create_cli_options.extend(["--workspace", workspace])

        if expire_at is not None:
            create_cli_options.extend([f"--expire-at={expire_at}"])

        if data is not None:
            data_file_temp = self.create_temporary_file(
                prefix="create-artifact", contents=data
            )
            create_cli_options.extend(["--data", str(data_file_temp)])

        if data_file is not None:
            if data is not None:
                raise ValueError(
                    "'data_file' parameter cannot be used with 'data'"
                )  # pragma: no cover
            create_cli_options.extend(["--data", str(data_file)])

        if files_to_upload is not None:
            create_cli_options.extend(["--upload", *files_to_upload])

        if upload_base_directory is not None:
            create_cli_options.extend(
                ["--upload-base-directory", upload_base_directory.as_posix()]
            )

        if stdin is not None:
            self.enterContext(self.patch_sys_stdin_read(yaml.safe_dump(stdin)))

        cli = self.create_cli(["artifact", "create", *create_cli_options])

        if assert_system_exit_code is None:
            stderr, stdout = self.capture_output(cli.execute)
        else:
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=assert_system_exit_code
            )

        return stderr, stdout, mocked_create_artifact

    def assert_create_artifact_success(
        self,
        files_to_upload: list[Path],
        upload_base_directory: Path | None = None,
        expire_at: dt.datetime | None = None,
    ) -> None:
        """
        Cli parse the command line and calls Debusine.create_artifact.

        Verify that the correct files_to_upload has been used.
        """
        artifact_category = ArtifactCategory.WORK_REQUEST_DEBUG_LOGS
        artifact_id = 2
        workspace = "test"

        # Mock Debusine.upload_files: will be used to verify the arguments
        # that debusine client used to try to upload the files
        upload_files_patcher = mock.patch.object(
            Debusine, "upload_files", autospec=True
        )
        upload_files_patcher.start()
        self.addCleanup(upload_files_patcher.stop)

        # Prepare dictionary with the paths in the artifact and the
        # FileModelRequest objects
        paths_in_artifact_to_file_models = {}
        for file_to_upload in files_to_upload:
            if file_to_upload.is_absolute():
                local_file_path = file_to_upload
                artifact_path = file_to_upload.name
            else:
                local_file_path = Path(os.getcwd()).joinpath(file_to_upload)
                artifact_path = str(file_to_upload.name)

            paths_in_artifact_to_file_models[artifact_path] = local_file_path

        create_artifact_return_value = create_artifact_response(
            scope=self.servers[self.default_server]["scope"],
            id=artifact_id,
            workspace=workspace,
            files_to_upload=list(paths_in_artifact_to_file_models.keys()),
        )

        (
            stderr_actual,
            stdout_actual,
            mocked_create_artifact,
        ) = self.create_artifact_via_cli(
            artifact_category,
            workspace=workspace,
            create_artifact_return_value=create_artifact_return_value,
            files_to_upload=list(map(str, files_to_upload)),
            upload_base_directory=upload_base_directory,
            expire_at=expire_at,
        )

        # Assert the call happened as expected
        artifact_request = self.assert_artifact_is_created(
            mocked_create_artifact,
            workspace,
            artifact_id,
            len(paths_in_artifact_to_file_models),
            stdout_actual,
            stderr_actual,
            expire_at,
        )
        self.assertEqual(artifact_request.category, artifact_category)
        self.assertEqual(
            artifact_request.files, paths_in_artifact_to_file_models
        )

    def test_create_artifact_absolute_file_paths(self) -> None:
        """Test create an artifact uploading files by absolute path."""
        self.assert_create_artifact_success(self.create_files_to_upload())

    def test_create_artifact_relative_file_path(self) -> None:
        """Test create an artifact uploading a file by relative to cwd."""
        file_path = self.create_temporary_file(directory=os.getcwd())

        file_name = file_path.name

        self.assert_create_artifact_success([Path(file_name)])

    def test_create_artifact_relative_file_path_in_subdirectory(self) -> None:
        """Test create an artifact uploading a file from a subdir of cwd."""
        temp_directory = self.create_temporary_directory(directory=os.getcwd())
        file_path = temp_directory / "file1.data"
        file_path.write_bytes(b"test")

        self.assert_create_artifact_success(
            [file_path.relative_to(os.getcwd())]
        )

    def test_create_artifact_upload_base_directory(self) -> None:
        """Create artifact with an abs file path in a abs path directory."""
        temp_directory = self.create_temporary_directory()

        (file := temp_directory / "file1.data").write_bytes(b"")

        self.assert_create_artifact_success([file], temp_directory)

    def test_create_artifact_expire_at(self) -> None:
        """Test create an artifact uploading files by absolute path."""
        expire_at = dt.datetime.now(tz=dt.UTC) + dt.timedelta(days=1)
        self.assert_create_artifact_success(
            self.create_files_to_upload(), expire_at=expire_at
        )

    def test_create_artifact_debusine_upload_files_fail(self) -> None:
        """Debusine.create_artifact fails: cannot connect to the server."""
        files_to_upload = self.create_files_to_upload()

        upload_files_patcher = mock.patch.object(
            Debusine, "upload_files", autospec=True
        )
        mocked_upload_files = upload_files_patcher.start()

        # Debusine.upload_files cannot connect to the server
        mocked_upload_files.side_effect = exceptions.ClientConnectionError
        self.addCleanup(upload_files_patcher.stop)

        workspace = "workspace"
        create_artifact_return_value = create_artifact_response(
            base_url=self.get_base_url(self.default_server),
            scope=self.servers[self.default_server]["scope"],
            id=2,
            workspace=workspace,
            files_to_upload=[p.name for p in files_to_upload],
        )

        stderr, _, _ = self.create_artifact_via_cli(
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
            workspace=workspace,
            create_artifact_return_value=create_artifact_return_value,
            files_to_upload=list(map(str, files_to_upload)),
            assert_system_exit_code=3,
        )
        self.assertTrue(stderr.startswith("Error connecting to debusine:"))

    def test_create_artifact_read_stdin(self) -> None:
        """Cli create an artifact reading the data from stdin."""
        artifact_id = 2
        artifact_category = ArtifactCategory.WORK_REQUEST_DEBUG_LOGS
        workspace = "default"
        data: dict[str, Any] = {}

        create_artifact_return_value = create_artifact_response(
            base_url=self.get_base_url(self.default_server),
            scope=self.servers[self.default_server]["scope"],
            id=artifact_id,
            workspace=workspace,
        )

        stderr, stdout, mocked_create_artifact = self.create_artifact_via_cli(
            artifact_category,
            workspace=workspace,
            create_artifact_return_value=create_artifact_return_value,
            stdin=data,
            data_file="-",
        )

        artifact_request = self.assert_artifact_is_created(
            mocked_create_artifact,
            workspace,
            artifact_id,
            0,
            stdout,
            stderr,
            data_from_stdin=True,
        )
        self.assertEqual(artifact_request.data.model_dump(), data)

    def test_create_artifact_invalid_data_file_type(self) -> None:
        """Cli try to create an artifact, data file does not contain a dict."""
        stderr, _, _ = self.create_artifact_via_cli(
            ArtifactCategory.SOURCE_PACKAGE,
            data=b"some-text",
            assert_system_exit_code=3,
        )

        self.assertEqual(
            stderr, "Error: data must be a dictionary. It is: str\n"
        )

    def test_create_artifact_invalid_yaml_in_data_file(self) -> None:
        """Cli try to create an artifact, data file contains invalid YAML."""
        stderr, _, _ = self.create_artifact_via_cli(
            ArtifactCategory.SOURCE_PACKAGE,
            data=b'"',
            assert_system_exit_code=3,
        )

        self.assertRegex(stderr, "^Error parsing YAML:")
        self.assertRegex(stderr, r"Fix the YAML data\n$")

    def test_create_artifact_cannot_read_data(self) -> None:
        """Cli try to create an artifact with a data file cannot be read."""
        data_file = "/tmp/debusine-test-file-does-not-exist"
        stderr, _, _ = self.create_artifact_via_cli(
            ArtifactCategory.SOURCE_PACKAGE,
            data_file=Path(data_file),
            assert_system_exit_code=2,
        )

        # argparse will show an error message containing the file path
        # (the rest of the message is up to argparse)
        self.assertRegex(stderr, data_file)

    def test_create_artifact_without_workspace_success(self) -> None:
        """Cli try to create an artifact with workspace=None."""
        artifact_id = 2
        expected_workspace = "default"
        data = {
            "name": "name",
            "version": "1.0",
            "type": "dpkg",
            "dsc_fields": {},
        }
        files_to_upload = self.create_files_to_upload()

        create_artifact_return_value = create_artifact_response(
            base_url=self.get_base_url(self.default_server),
            scope=self.servers[self.default_server]["scope"],
            id=artifact_id,
            workspace=expected_workspace,
            files_to_upload=[path.name for path in files_to_upload],
        )
        artifact_category = ArtifactCategory.SOURCE_PACKAGE
        with mock.patch.object(Debusine, "upload_files", autospec=True):
            (
                stderr_actual,
                stdout_actual,
                mocked_create_artifact,
            ) = self.create_artifact_via_cli(
                artifact_category,
                data=yaml.safe_dump(data).encode("utf-8"),
                create_artifact_return_value=create_artifact_return_value,
                files_to_upload=[str(path) for path in files_to_upload],
            )

        self.assert_artifact_is_created(
            mocked_create_artifact,
            expected_workspace,
            artifact_id,
            len(files_to_upload),
            stdout_actual,
            stderr_actual,
        )

    def test_create_artifact_workspace_not_found(self) -> None:
        """Cli try to create an artifact for a non-existing workspace."""
        workspace_name = "does-not-exist"
        title_error = f'Workspace "{workspace_name}" cannot be found'

        create_artifact_side_effect = DebusineError(title=title_error)

        stderr_actual, stdout_actual, _ = self.create_artifact_via_cli(
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
            workspace=workspace_name,
            create_artifact_side_effect=create_artifact_side_effect,
            assert_system_exit_code=3,
        )

        stdout_expected = yaml.safe_dump(
            {"result": "failure", "error": {"title": title_error}},
            sort_keys=False,
        )

        self.assertEqual(stderr_actual, "")
        self.assertEqual(stdout_actual, stdout_expected)

    def test_create_artifact_with_two_files_of_same_name(self) -> None:
        """Cannot create an artifact with two files having the same path."""
        file_name = "file1"
        # Create a file to be uploaded in a directory
        directory_to_upload1 = self.create_temporary_directory()
        file_in_dir1 = Path(directory_to_upload1) / Path(file_name)
        file_in_dir1.write_bytes(b"test")

        # And create another file, same name, in another directory
        directory_to_upload2 = self.create_temporary_directory()
        file_in_dir2 = Path(directory_to_upload2) / Path(file_name)
        file_in_dir2.write_bytes(b"test")

        paths_to_upload = [str(file_in_dir1), str(file_in_dir2)]

        stderr, _, _ = self.create_artifact_via_cli(
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
            workspace="some-workspace",
            files_to_upload=paths_to_upload,
            assert_system_exit_code=3,
        )

        self.assertEqual(
            stderr,
            f"Cannot create artifact: File with the same path ({file_name}) "
            f"is already in the artifact "
            f'("{file_in_dir1}" and "{file_in_dir2}")\n',
        )

    def test_create_artifact_invalid_category(self) -> None:
        """Cli try to create an artifact with a category that does not exist."""
        self.create_artifact_via_cli(
            "missing:does-not-exist", assert_system_exit_code=2
        )


class CliImportDebianArtifactTests(BaseCliTests):
    """
    Tests for the debian artifact import CLI.

    Mocking at the upload_artifact level.

    Remote sources are tested through the DGet test suite, so not repeated here.
    """

    def create_dsc(self, directory: Path | None = None) -> list[Path]:
        """
        Create a simple Debian source package in a temporary directory.

        Return a list of the created Paths. The first item is the dsc.
        """
        if not directory:
            directory = self.create_temporary_directory()
        orig = directory / "pkg_1.0.orig.tar.gz"
        orig.write_bytes(b"I'm a tarball")
        dsc = directory / "pkg_1.0.dsc"
        self.write_dsc_file(dsc, [orig], version="1.0")
        return [dsc, orig]

    def create_upload(
        self,
        source: bool,
        binary: bool,
        extra_dsc: bool = False,
        binnmu: bool = False,
    ) -> list[Path]:
        """
        Create a simple Debian upload in a temporary directory.

        Return a list of the created Paths. The first item is the changes.
        """
        directory = self.create_temporary_directory()
        files = []
        if source:
            files += self.create_dsc(directory)
        if extra_dsc:
            extra = directory / "foo_1.0.dsc"
            self.write_dsc_file(extra, [], version="1.0")
            files.append(extra)
        if binary:
            deb_version = "1.0"
            if binnmu:
                deb_version += "+b1"
            deb = directory / f"pkg_{deb_version}_all.deb"
            self.write_deb_file(deb, source_version="1.0")
            files.append(deb)
        changes = directory / "pkg.changes"
        self.write_changes_file(changes, files, version="1.0", binnmu=binnmu)
        return [changes, *files]

    def patch_upload_artifact(self) -> MagicMock:
        """
        Patch Debusine.upload_artifact.

        Return different RemoteArtifacts for each request.
        """
        patcher_upload_artifact = mock.patch.object(
            Debusine, "upload_artifact", autospec=True
        )
        mocked_upload_artifact = patcher_upload_artifact.start()

        id_counter = count(10)

        def upload_artifact(
            _self: Debusine,
            local_artifact: LocalArtifact[Any],  # noqa: ARG001, U100
            *,
            workspace: str | None,
            expire_at: dt.datetime | None,  # noqa: ARG001, U100
        ) -> RemoteArtifact:
            if workspace is None:  # pragma: no cover
                workspace = "System"
            artifact_id = next(id_counter)
            return create_remote_artifact(
                base_url=self.get_base_url(self.default_server),
                scope=self.servers[self.default_server]["scope"],
                id=artifact_id,
                workspace=workspace,
            )

        mocked_upload_artifact.side_effect = upload_artifact

        self.addCleanup(patcher_upload_artifact.stop)
        return mocked_upload_artifact

    def patch_relation_create(self) -> MagicMock:
        """Patch Debusine.relation_create."""
        patcher_relation_create = mock.patch.object(
            Debusine, "relation_create", autospec=True
        )
        mocked_relation_create = patcher_relation_create.start()
        self.addCleanup(patcher_relation_create.stop)
        return mocked_relation_create

    def assert_artifacts_are_uploaded(
        self,
        mocked_upload_artifact: MagicMock,
        expected_workspace: str,
        expected_artifacts: int,
        expected_files: int,
        stdout: str,
        stderr: str,
        extended_artifacts: list[int] | None = None,
        related_artifacts: list[int] | None = None,
        expected_expire_at: dt.datetime | None = None,
    ) -> list[LocalArtifact[Any]]:
        """
        Assert that mocked_upload_artifact uploaded expected_artifacts only.

        In stdout we expect that CLI printed the uploaded artifacts.

        Return LocalArtifacts that would have been uploaded.

        :param mocked_upload_artifact: mock of upload_artifact.
        :param expected_workspace: expected created workspace.
        :param expected_artifacts: expected number of artifacts created.
        :param expected_files: expected number of files uploaded.
        :param extended_artifacts: expected artifact IDs extending the main
            artifact.
        :param related_artifacts: expected artifact IDs relating to the main
            artifact.
        :param stdout: contains Cli.execute() output to verify expected output.
        :param stderr: contains Cli.execute() output to verify expected output.
        """
        self.assertEqual(mocked_upload_artifact.call_count, expected_artifacts)

        scope = self.servers[self.default_server]["scope"]
        artifact_url = (
            f"{quote(scope)}/{quote(expected_workspace)}/artifact/10/"
        )

        expected = {
            "result": "success",
            "message": f"New artifact created: {artifact_url}",
            "artifact_id": 10,
        }
        if extended_artifacts:
            expected["extends"] = [
                {"artifact_id": id_} for id_ in extended_artifacts
            ]
        if related_artifacts:
            expected["relates_to"] = [
                {"artifact_id": id_} for id_ in related_artifacts
            ]
        expected_str = yaml.safe_dump(
            expected,
            sort_keys=False,
            width=math.inf,
        )

        self.assertEqual(stdout, expected_str)
        self.assertEqual(stderr, "")

        local_artifacts = []
        uploaded_files = set()
        for args, kwargs in mocked_upload_artifact.call_args_list:
            self.assertEqual(len(args), 2)
            _, local_artifact = args
            self.assertEqual(kwargs["workspace"], expected_workspace)
            self.assertEqual(kwargs["expire_at"], expected_expire_at)
            self.assertIsInstance(local_artifact, LocalArtifact)
            for file in local_artifact.files:
                uploaded_files.add(file)
            local_artifacts.append(local_artifact)

        self.assertEqual(len(uploaded_files), expected_files)

        return local_artifacts

    def import_debian_artifact(
        self,
        *,
        workspace: str | None = None,
        upload: str | os.PathLike[str] | None = None,
        expire_at: dt.datetime | None = None,
        upload_artifact_side_effect: (
            BaseException | type[BaseException] | None
        ) = None,
        assert_system_exit_code: int | None = None,
        debusine_error: dict[str, Any] | None = None,
        exception_str: str | None = None,
    ) -> tuple[str, str, MagicMock]:
        """
        Call CLI import-debian-artifact with the correct parameters.

        Patch Debusine.upload_artifact before executing the command.

        :param workspace: add ["--workspace", workspace] options.
        :param upload: file path / URL to upload.
        :param expire_at: datetime (for --expire-at option).
        :param upload_artifact_side_effect: side effect for the
          Debusine.create_artifact mock.
        :param assert_system_exit_code: assert that the import-debian-artifact
          command raises SystemExit(X).

        Return tuple with stderr, stdout and the patched
        Debusine.upload_artifact
        """
        mocked_upload_artifact = self.patch_upload_artifact()

        if upload_artifact_side_effect is not None:
            mocked_upload_artifact.side_effect = upload_artifact_side_effect

        args = []

        if workspace is not None:
            args += ["--workspace", workspace]

        if expire_at is not None:
            args += ["--expire-at", str(expire_at)]

        args.append(str(upload))

        cli = self.create_cli(["artifact", "import-debian", *args])

        if debusine_error is not None:
            exception = self.assertShowsError(cli.execute)
            self.assertDebusineError(exception, debusine_error)
            stdout, stderr = "", ""
        elif exception_str is not None:
            exception = self.assertShowsError(cli.execute)
            self.assertEqual(str(exception), exception_str)
            stdout, stderr = "", ""
        elif assert_system_exit_code is not None:
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=assert_system_exit_code
            )
        else:
            stderr, stdout = self.capture_output(cli.execute)

        return stderr, stdout, mocked_upload_artifact

    def assert_import_dsc(
        self,
        dsc: Path,
        files: list[Path],
        workspace: str = "test",
        expire_at: dt.datetime | None = None,
    ) -> None:
        """
        Call import-debian-artifact to import a source package.

        Verify that the right artifacts are created and files uploaded.

        Return the uploaded artifacts.
        """
        # Make the request
        (
            stderr_actual,
            stdout_actual,
            mocked_upload_artifact,
        ) = self.import_debian_artifact(
            workspace=workspace,
            upload=str(dsc),
            expire_at=expire_at,
        )

        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_upload_artifact,
            expected_workspace=workspace,
            expected_artifacts=1,
            expected_files=len(files),
            stdout=stdout_actual,
            stderr=stderr_actual,
            expected_expire_at=expire_at,
        )
        self.assertEqual(len(artifacts), 1)
        self.assertSourcePackage(artifacts[0])

    def assertSourcePackage(self, artifact: LocalArtifact[Any]) -> None:
        """Assert that artifact is a .dsc with expected metadata."""
        self.assertEqual(artifact.category, ArtifactCategory.SOURCE_PACKAGE)
        self.assertEqual(artifact.data.name, "hello-traditional")
        self.assertEqual(artifact.data.type, "dpkg")
        self.assertEqual(artifact.data.version, "1.0")
        self.assertEqual(
            artifact.data.dsc_fields["Source"], "hello-traditional"
        )
        self.assertEqual(
            set(artifact.files), {"pkg_1.0.dsc", "pkg_1.0.orig.tar.gz"}
        )

    def assertBinaryPackage(
        self, artifact: LocalArtifact[Any], binnmu: bool = False
    ) -> None:
        """Assert that artifact is a .deb with expected metadata."""
        self.assertEqual(artifact.category, ArtifactCategory.BINARY_PACKAGE)
        version = "1.0"
        if binnmu:
            version += "+b1"
        expected_deb_fields = {
            "Package": "pkg",
            "Version": version,
            "Architecture": "all",
            "Maintainer": "Example Maintainer <example@example.org>",
            "Description": "Example description",
        }
        if binnmu:
            expected_deb_fields["Source"] = "pkg (1.0)"
        expected = {
            "srcpkg_name": "pkg",
            "srcpkg_version": "1.0",
            "deb_fields": expected_deb_fields,
            "deb_control_files": ["control"],
        }
        self.assertEqual(artifact.data.model_dump(), expected)
        self.assertEqual(set(artifact.files), {f"pkg_{version}_all.deb"})

    def assertUpload(
        self,
        artifact: LocalArtifact[Any],
        file_names: set[str],
        binnmu: bool = False,
    ) -> None:
        """Assert that artifact is an Upload with expected metadata."""
        self.assertEqual(artifact.category, ArtifactCategory.UPLOAD)
        self.assertEqual(artifact.data.type, "dpkg")
        expected = "hello-traditional"
        if binnmu:
            expected += " (1.0)"
        self.assertEqual(artifact.data.changes_fields["Source"], expected)
        self.assertEqual(set(artifact.files), file_names)

    def test_import_debian_artifact_dsc(self) -> None:
        """Test import a .dsc from a local path."""
        files = self.create_dsc()
        self.assert_import_dsc(files[0], files)

    def test_import_debian_artifact_expire_at(self) -> None:
        """Test import a .dsc from a local path, with expiry."""
        expire_at = dt.datetime.now(tz=dt.UTC) + dt.timedelta(days=1)
        files = self.create_dsc()
        self.assert_import_dsc(files[0], files, expire_at=expire_at)

    def test_import_debian_artifact_workspace(self) -> None:
        """Test import a .dsc from a local path, with a workspace."""
        files = self.create_dsc()
        self.assert_import_dsc(files[0], files, workspace="different")

    def test_import_debian_artifact_source_changes(self) -> None:
        """Test import a source upload from a local path."""
        files = self.create_upload(source=True, binary=False)
        workspace = "test"
        mocked_relation_create = self.patch_relation_create()
        stderr, stdout, mocked_upload_artifact = self.import_debian_artifact(
            upload=str(files[0]),
            workspace=workspace,
        )

        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_upload_artifact,
            expected_workspace=workspace,
            expected_artifacts=2,
            expected_files=len(files),
            stdout=stdout,
            stderr=stderr,
            extended_artifacts=[11],
            related_artifacts=[11],
        )
        self.assertUpload(artifacts[0], {file.name for file in files})
        self.assertSourcePackage(artifacts[1])
        mocked_relation_create.assert_has_calls(
            [
                mock.call(
                    mock.ANY,
                    artifact_id=10,
                    target_id=11,
                    relation_type="extends",
                ),
                mock.call(
                    mock.ANY,
                    artifact_id=10,
                    target_id=11,
                    relation_type="relates-to",
                ),
            ]
        )

    def test_import_debian_artifact_binary_changes(self) -> None:
        """Test import a binary upload from a local path."""
        files = self.create_upload(source=False, binary=True)
        workspace = "test"
        mocked_relation_create = self.patch_relation_create()
        stderr, stdout, mocked_upload_artifact = self.import_debian_artifact(
            upload=str(files[0]),
            workspace=workspace,
        )

        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_upload_artifact,
            expected_workspace=workspace,
            expected_artifacts=2,
            expected_files=len(files),
            stdout=stdout,
            stderr=stderr,
            extended_artifacts=[11],
            related_artifacts=[11],
        )
        self.assertUpload(artifacts[0], {file.name for file in files})
        self.assertBinaryPackage(artifacts[1])
        mocked_relation_create.assert_has_calls(
            [
                mock.call(
                    mock.ANY,
                    artifact_id=10,
                    target_id=11,
                    relation_type="extends",
                ),
                mock.call(
                    mock.ANY,
                    artifact_id=10,
                    target_id=11,
                    relation_type="relates-to",
                ),
            ]
        )

    def test_import_debian_artifact_binnmu_changes(self) -> None:
        """Test import a binnmu binary upload from a local path."""
        files = self.create_upload(source=False, binary=True, binnmu=True)
        workspace = "test"
        self.patch_relation_create()
        stderr, stdout, mocked_upload_artifact = self.import_debian_artifact(
            upload=str(files[0]),
            workspace=workspace,
        )

        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_upload_artifact,
            expected_workspace=workspace,
            expected_artifacts=2,
            expected_files=len(files),
            stdout=stdout,
            stderr=stderr,
            extended_artifacts=[11],
            related_artifacts=[11],
        )
        self.assertUpload(
            artifacts[0], {file.name for file in files}, binnmu=True
        )
        self.assertBinaryPackage(artifacts[1], binnmu=True)

    def test_import_debian_artifact_mixed_changes(self) -> None:
        """Test import a source+binary upload from a local path."""
        files = self.create_upload(source=True, binary=True)
        workspace = "test"
        self.patch_relation_create()
        stderr, stdout, mocked_upload_artifact = self.import_debian_artifact(
            upload=str(files[0]),
            workspace=workspace,
        )

        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_upload_artifact,
            expected_workspace=workspace,
            expected_artifacts=3,
            expected_files=len(files),
            stdout=stdout,
            stderr=stderr,
            extended_artifacts=[11, 12],
            related_artifacts=[11, 12],
        )
        self.assertUpload(artifacts[0], {file.name for file in files})
        self.assertBinaryPackage(artifacts[1])
        self.assertSourcePackage(artifacts[2])

    def test_import_debian_artifact_deb(self) -> None:
        """import-debian-artifact imports a .deb directly."""
        directory = self.create_temporary_directory()
        workspace = "test"
        deb = directory / "pkg_1.0_all.deb"
        self.write_deb_file(deb)
        stderr, stdout, mocked_create_artifact = self.import_debian_artifact(
            upload=deb,
            workspace=workspace,
        )
        # Assert the call happened as expected
        artifacts = self.assert_artifacts_are_uploaded(
            mocked_create_artifact,
            expected_workspace=workspace,
            expected_artifacts=1,
            expected_files=1,
            stdout=stdout,
            stderr=stderr,
        )
        self.assertBinaryPackage(artifacts[0])

    def test_import_debian_artifact_fails_to_connect(self) -> None:
        """Debusine.upload_artifact fails: cannot connect to the server."""
        files = self.create_dsc()

        stderr, _, _ = self.import_debian_artifact(
            workspace="test",
            upload=str(files[0]),
            upload_artifact_side_effect=exceptions.ClientConnectionError,
            assert_system_exit_code=3,
        )
        self.assertTrue(stderr.startswith("Error connecting to debusine:"))

    def test_import_debian_artifact_multiple_dsc(self) -> None:
        """import-debian-artifact fails when given two .dscs."""
        files = self.create_upload(source=True, binary=False, extra_dsc=True)
        stderr, _, _ = self.import_debian_artifact(
            upload=files[0], assert_system_exit_code=3
        )

        self.assertRegex(
            stderr, r"^Expecting exactly 1 \.dsc in source package, found 2$"
        )

    def test_import_debian_artifact_binary_changes_without_debs(self) -> None:
        """import-debian-artifact fails with a binary upload without debs."""
        changes = self.create_temporary_file(suffix=".changes")
        changes.write_bytes(
            textwrap.dedent(
                """\
                Format: 3.0 (quilt)
                Source: hello-traditional
                Binary: hello-traditional
                Version: 2.10-5
                Maintainer: Santiago Vila <sanvila@debian.org>
                Homepage: http://www.gnu.org/software/hello/
                Standards-Version: 4.3.0
                Architecture: all
                Package-List:
                 hello-traditional deb devel optional arch=any
                Files:
                """
            ).encode("utf-8")
        )
        stderr, _, _ = self.import_debian_artifact(
            upload=changes, assert_system_exit_code=3
        )

        self.assertRegex(
            stderr,
            r"^Expecting at least one \.deb per arch in binary packages$",
        )

    def test_import_debian_artifact_unknown_extension(self) -> None:
        """import-debian-artifact fails when given a .foo file."""
        file = self.create_temporary_file(suffix=".foo")
        self.import_debian_artifact(
            upload=file,
            exception_str=(
                "Only source packages (.dsc), binary packages (.deb), "
                "and source/binary uploads (.changes) can be directly "
                f"imported with this command. {file} is not supported."
            ),
        )

    def test_import_debian_artifact_missing_file(self) -> None:
        """import-debian-artifact fails when a named file is missing."""
        file = "/tmp/nonexistent.dsc"
        self.import_debian_artifact(
            upload=file,
            exception_str=(
                "[Errno 2] No such file or directory: '/tmp/nonexistent.dsc'"
            ),
        )

    def test_import_debian_artifact_missing_referenced_file(self) -> None:
        """import-debian-artifact fails when a referenced file is missing."""
        files = self.create_dsc()
        files[1].unlink()
        self.import_debian_artifact(
            upload=files[0], exception_str=f'"{files[1]}" does not exist'
        )

    def test_import_debian_artifact_workspace_not_found(self) -> None:
        """import-debian-artifact fails when the named workspace is missing."""
        workspace_name = "does-not-exist"
        title_error = f'Workspace "{workspace_name}" cannot be found'
        files = self.create_dsc()

        self.import_debian_artifact(
            workspace="test",
            upload=str(files[0]),
            upload_artifact_side_effect=DebusineError(title=title_error),
            debusine_error={"title": title_error},
        )


class CliShowArtifactTests(BaseCliTests):
    """Tests for Cli show-artifact."""

    def test_show_artifact(self) -> None:
        """Test show-artifact shows the information."""
        artifact_id = 11
        debusine = self.patch_build_debusine_object()
        debusine.return_value.artifact_get.return_value = (
            create_artifact_response(id=artifact_id)
        )
        cmd = self.create_command(["artifact", "show", str(artifact_id)])
        assert isinstance(cmd, DebusineCommand)

        with mock.patch(
            "debusine.client.commands.base.DebusineCommand._api_call_or_fail",
            side_effect=cmd._api_call_or_fail,
        ) as api_call_or_fail_mocked:
            stderr, stdout = self.capture_output(cmd.run)

        api_call_or_fail_mocked.assert_called_once_with()
        debusine.return_value.artifact_get.assert_called_once_with(
            int(artifact_id)
        )

        self.assertEqual(stderr, "")

        _, expected_output = self.capture_output(
            DebusineCommand._print_yaml,
            [
                debusine.return_value.artifact_get.return_value.model_dump(
                    mode="json"
                )
            ],
        )

        self.assertEqual(stdout, expected_output)


class CliShowArtifactRelationsTests(BaseCliTests):
    """Tests for Cli show-artifact-relations."""

    def create_relation_response(self, **kwargs: Any) -> RelationResponse:
        """Return a RelationResponse. Use defaults for certain fields."""
        defaults: dict[str, Any] = {
            "id": 1,
            "artifact": 10,
            "target": 11,
            "type": "relates-to",
        }
        defaults.update(kwargs)

        return RelationResponse(**defaults)

    def create_relations_response(
        self, *args: RelationResponse
    ) -> RelationsResponse:
        """Return a RelationsResponse containing args."""
        return RelationsResponse(args)

    def test_show_artifact_relations(self) -> None:
        """Test show-artifact-relations shows all relations."""
        artifact_id = 11
        debusine = self.patch_build_debusine_object()
        debusine.return_value.relation_list.return_value = (
            self.create_relations_response(
                self.create_relation_response(
                    id=1, artifact=artifact_id, target=15
                ),
                self.create_relation_response(
                    id=2, artifact=artifact_id, target=16
                ),
            )
        )
        cmd = self.create_command(
            ["artifact", "show-relations", str(artifact_id)]
        )
        assert isinstance(cmd, DebusineCommand)

        with mock.patch(
            "debusine.client.commands.base.DebusineCommand._api_call_or_fail",
            side_effect=cmd._api_call_or_fail,
        ) as api_call_or_fail_mocked:
            stderr, stdout = self.capture_output(cmd.run)

        api_call_or_fail_mocked.assert_called_once_with()
        debusine.return_value.relation_list.assert_called_once_with(
            artifact_id=int(artifact_id)
        )

        self.assertEqual(stderr, "")

        output = RelationsResponse(yaml.safe_load(stdout))
        expected = RelationsResponse(
            [
                RelationResponse(
                    artifact=11,
                    target=15,
                    type=RelationType.RELATES_TO,
                    id=1,
                ),
                RelationResponse(
                    artifact=11,
                    target=16,
                    type=RelationType.RELATES_TO,
                    id=2,
                ),
            ]
        )

        self.assertEqual(output, expected)

    def test_show_artifact_relations_reverse(self) -> None:
        """Test show-artifact-relations shows reverse relations."""
        artifact_id = 11
        debusine = self.patch_build_debusine_object()
        debusine.return_value.relation_list.return_value = (
            self.create_relations_response(
                self.create_relation_response(
                    id=3, artifact=8, target=artifact_id
                ),
                self.create_relation_response(
                    id=4, artifact=9, target=artifact_id
                ),
            )
        )
        cmd = self.create_command(
            ["artifact", "show-relations", "--reverse", str(artifact_id)]
        )
        assert isinstance(cmd, DebusineCommand)

        with mock.patch(
            "debusine.client.commands.base.DebusineCommand._api_call_or_fail",
            side_effect=cmd._api_call_or_fail,
        ) as api_call_or_fail_mocked:
            stderr, stdout = self.capture_output(cmd.run)

        api_call_or_fail_mocked.assert_called_once_with()
        debusine.return_value.relation_list.assert_called_once_with(
            target_id=int(artifact_id)
        )

        self.assertEqual(stderr, "")

        output = RelationsResponse(yaml.safe_load(stdout))
        expected = RelationsResponse(
            [
                RelationResponse(
                    artifact=8,
                    target=11,
                    type=RelationType.RELATES_TO,
                    id=3,
                ),
                RelationResponse(
                    artifact=9,
                    target=11,
                    type=RelationType.RELATES_TO,
                    id=4,
                ),
            ]
        )
        self.assertEqual(output, expected)


class CliDownloadArtifact(BaseCliTests):
    """Tests for the Cli functionality related to downloading an artifact."""

    def test_download_artifact(self) -> None:
        """download-artifact call debusine.download_artifact()."""
        artifact_id = str(10)
        destination = Path.cwd()

        debusine = self.patch_build_debusine_object()

        for args, tarball in (
            (["artifact", "download", artifact_id, "--tarball"], True),
            (["artifact", "download", artifact_id], False),
            (["-s", "artifact", "download", artifact_id, "--tarball"], True),
            (["-s", "artifact", "download", artifact_id], False),
        ):
            with self.subTest(args=args):
                cli = self.create_cli(args)

                cli.execute()

                debusine.return_value.download_artifact.assert_called_with(
                    int(artifact_id), destination=destination, tarball=tarball
                )

    def test_download_artifact_target_directory_no_write_access(self) -> None:
        """_download_artifact target_directory no write access."""
        # Similar as other tests: avoid building the debusine object
        # (not relevant for this test)
        self.patch_build_debusine_object()

        target_directory = self.create_temporary_directory()
        os.chmod(target_directory, 0)

        cli = self.create_cli(
            [
                "artifact",
                "download",
                "10",
                "--target-directory",
                str(target_directory),
            ],
        )

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(stderr, f"Error: Cannot write to {target_directory}\n")
        self.assertEqual(stdout, "")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Queue and worker pool summaries."""

import abc
import logging
from collections import defaultdict

import pydantic
from django.db.models import Count

import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest, Worker
from debusine.tasks.models import WorkerType

log = logging.getLogger(__name__)


class PydanticBase(pydantic.BaseModel, abc.ABC):
    """Stricter BaseModel."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")


class WorkerSummary(PydanticBase):
    """Worker pool count."""

    registered: int = 0
    connected: int = 0
    idle: int = 0
    busy: int = 0


class TaskQueueSummary(PydanticBase):
    """Work Request queue length."""

    pending: int = 0
    running: int = 0


class QueueStatus(PydanticBase):
    """Worker queue lengths and pool sizes."""

    tasks: dict[TaskTypes, TaskQueueSummary] = {}

    # by worker type
    workers: dict[WorkerType, WorkerSummary] = {}
    worker_tasks: dict[WorkerType, TaskQueueSummary] = {}

    # by architecture
    external_workers_arch: dict[str, WorkerSummary] = {}
    external_workers_native_arch: dict[str, WorkerSummary] = {}
    worker_tasks_arch: dict[str | None, TaskQueueSummary] = {}


class WorkerStatus:
    """Calculate the size of queues and the worker pool."""

    @classmethod
    def get_status(cls) -> QueueStatus:
        """Count number of tasks and workers by type and architecture."""
        result = QueueStatus()

        result.tasks = cls._count_queued_tasks_by_type()
        result.workers = cls._count_workers_by_type()
        result.external_workers_arch = (
            cls._count_external_workers_by_architecture()
        )
        result.external_workers_native_arch = (
            cls._count_external_workers_by_native_architecture()
        )
        result.worker_tasks = cls._count_task_by_worker_type(result.tasks)

        # Add architectures with queued work requests
        for (
            arch,
            queued_work_requests,
        ) in cls._count_queued_worker_work_requests_by_architecture().items():
            if arch and arch not in result.external_workers_arch:
                result.external_workers_arch[arch] = WorkerSummary()
            result.worker_tasks_arch[arch] = queued_work_requests

        # Add architectures with online workers, and no queued work requests
        # (those with queued work requests were already added)
        for arch in result.external_workers_arch:
            if arch not in result.worker_tasks_arch:
                result.worker_tasks_arch[arch] = TaskQueueSummary()

        return result

    @classmethod
    def _count_queued_tasks_by_type(cls) -> dict[TaskTypes, TaskQueueSummary]:
        """Return summary of queued tasks by task type."""
        result: dict[TaskTypes, TaskQueueSummary] = {
            task_type: TaskQueueSummary() for task_type in TaskTypes
        }

        for row in (
            WorkRequest.objects.pending()
            .order_by()
            .values("task_type")
            .annotate(count=Count("id"))
        ):
            result[row["task_type"]].pending = row["count"]

        for row in (
            WorkRequest.objects.running()
            .values("task_type")
            .annotate(count=Count("id"))
        ):
            result[row["task_type"]].running = row["count"]

        return result

    @staticmethod
    def _count_task_by_worker_type(
        tasks: dict[TaskTypes, TaskQueueSummary],
    ) -> dict[WorkerType, TaskQueueSummary]:
        """Re-count queued tasks by worker type."""
        result: dict[WorkerType, TaskQueueSummary] = {
            worker_type: TaskQueueSummary()
            for worker_type in WorkerType
            if worker_type != WorkerType.SCHEDULER
        }
        for task_type, worker_type in (
            (TaskTypes.WORKER, WorkerType.EXTERNAL),
            (TaskTypes.SERVER, WorkerType.CELERY),
            (TaskTypes.SIGNING, WorkerType.SIGNING),
        ):
            for field_name in TaskQueueSummary.model_fields:
                setattr(
                    result[worker_type],
                    field_name,
                    getattr(result[worker_type], field_name)
                    + getattr(tasks[task_type], field_name),
                )
        return result

    @classmethod
    def _count_queued_worker_work_requests_by_architecture(
        cls,
    ) -> dict[str | None, TaskQueueSummary]:
        """
        Return summary of queued WORKER work requests by architecture.

        Categorize all WORKER tasks that don't have a specified architecture as
        None.
        """
        # TODO: this query can probably be optimized to directly aggregate on
        # the presence of tags by prefix
        pending_tagsets = (
            (WorkRequest.objects.pending() | WorkRequest.objects.running())
            .filter(task_type=TaskTypes.WORKER)
            .values("status", "scheduler_tags_required")
            .annotate(count=Count("id"))
            .order_by()
        )
        result: defaultdict[str | None, TaskQueueSummary] = defaultdict(
            TaskQueueSummary
        )
        for entry in pending_tagsets:
            matching_arches = (
                tag.removeprefix(wtags.BUILD_ARCH_PREFIX)
                for tag in entry["scheduler_tags_required"]
                if tag.startswith(wtags.BUILD_ARCH_PREFIX)
            )
            arch = next(matching_arches, None)
            if entry["status"] == WorkRequest.Statuses.PENDING:
                result[arch].pending += entry["count"]
            else:
                result[arch].running += entry["count"]

        return dict(result)

    @staticmethod
    def _count_workers_by_type() -> dict[WorkerType, WorkerSummary]:
        """Return number of connected external workers by type."""
        result: dict[WorkerType, WorkerSummary] = {
            worker_type: WorkerSummary()
            for worker_type in WorkerType
            if worker_type != WorkerType.SCHEDULER
        }

        for worker in Worker.objects.filter(token__enabled=True).exclude(
            worker_type=WorkerType.SCHEDULER
        ):
            worker_type = WorkerType[worker.worker_type.upper()]
            result[worker_type].registered += 1
            if worker.connected():
                result[worker_type].connected += 1
                if worker.is_busy():
                    result[worker_type].busy += 1
                else:
                    result[worker_type].idle += 1

        return result

    @staticmethod
    def _count_external_workers_by_architecture() -> dict[str, WorkerSummary]:
        """
        Return number of connected external workers by architecture.

        The same worker might be able to execute tasks for multiple
        architectures.
        """
        result: defaultdict[str, WorkerSummary] = defaultdict(WorkerSummary)

        for worker in Worker.objects.filter(
            token__enabled=True, worker_type=WorkerType.EXTERNAL
        ):
            for arch in worker.metadata().get("system:architectures", []):
                result[arch].registered += 1
                if worker.connected():
                    result[arch].connected += 1
                    if worker.is_busy():
                        result[arch].busy += 1
                    else:
                        result[arch].idle += 1

        return dict(result)

    @staticmethod
    def _count_external_workers_by_native_architecture() -> dict[
        str, WorkerSummary
    ]:
        """
        Return number of connected external workers by native architecture.

        The same worker might be able to execute tasks for multiple
        architectures, but this only returns the worker's native
        architecture.
        """
        result: defaultdict[str, WorkerSummary] = defaultdict(WorkerSummary)

        for worker in Worker.objects.filter(
            token__enabled=True,
            worker_type=WorkerType.EXTERNAL,
        ):
            worker_metadata = worker.metadata()
            # TODO: Drop system:host_architecture fallback once all workers
            # have been updated.
            arch = worker_metadata.get(
                "system:native_architecture",
                worker_metadata.get("system:host_architecture"),
            )
            if not arch:
                log.warning(
                    "Worker %s missing system:native_architecture", worker.name
                )
                continue
            result[arch].registered += 1
            if worker.connected():
                result[arch].connected += 1
                if worker.is_busy():
                    result[arch].busy += 1
                else:
                    result[arch].idle += 1

        return dict(result)



# Generated by Django 4.2.28 on 2026-06-04 16:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0059_workflowtemplate_extra_groups'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionRole',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'role',
                    models.CharField(
                        choices=[('owner', 'Owner'), ('viewer', 'Viewer')],
                        max_length=16,
                    ),
                ),
                (
                    'group',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='collection_roles',
                        to='db.group',
                    ),
                ),
                (
                    'resource',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='roles',
                        to='db.collection',
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name='collectionrole',
            constraint=models.UniqueConstraint(
                fields=('resource', 'group', 'role'),
                name='db_collectionrole_unique_resource_group_role',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for collections."""

from functools import cached_property
from typing import override

from django.urls import reverse
from django.utils.html import format_html

from debusine.db.models import Collection, CollectionItem
from debusine.db.models.collections import CollectionRelation
from debusine.web.views import ui_shortcuts
from debusine.web.views.places import IndexPlace, ResourcePlace
from debusine.web.views.ui.base import UI


class CollectionUI(UI[Collection]):
    """UI helpers for Collection instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this collection."""
        place = ResourcePlace(
            title=f"{self.instance.name}@{self.instance.category} collection",
            breadcrumb=self.instance.name,
            url=self.instance.get_absolute_url(),
            description=f"{self.instance.name}@{self.instance.category}"
            " collection",
        )
        place.parent = self.instance.workspace.ui(
            self.request
        ).place_collection_list
        return place

    @cached_property
    def place_search(self) -> IndexPlace:
        """Return a place to search this collection."""
        return IndexPlace(
            title=f"Search {self.instance.name}@{self.instance.category}",
            url=self.instance.get_absolute_url_search(),
            description=f"{self.instance} collection search",
            parent=self.instance.workspace.ui(
                self.request
            ).place_collection_list,
        )

    @cached_property
    def place_update(self) -> IndexPlace:
        """Return a place to configure this collection."""
        return IndexPlace(
            title="Configure properties",
            url=reverse(
                "workspaces:collections:update",
                kwargs={
                    "wname": self.instance.workspace.name,
                    "cname": self.instance.name,
                    "ccat": self.instance.category,
                },
            ),
            parent=self.place,
        )

    @cached_property
    def place_delete(self) -> IndexPlace:
        """Return a place to delete this collection."""
        return IndexPlace(
            title="Delete",
            url=reverse(
                "workspaces:collections:delete",
                kwargs={
                    "wname": self.instance.workspace.name,
                    "cname": self.instance.name,
                    "ccat": self.instance.category,
                },
            ),
            parent=self.place,
        )

    @cached_property
    def place_collection_relation_configure(self) -> IndexPlace:
        """Return a place to configure this collection."""
        return IndexPlace(
            title="Configure relations",
            url=reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.instance.workspace.name,
                    "cname": self.instance.name,
                    "ccat": self.instance.category,
                },
            ),
        )

    @cached_property
    def place_collection_relation_list(self) -> ResourcePlace:
        """Return a place to list similar relations for this collection."""
        collection = self.instance

        url = reverse(
            "workspaces:collections:collection_relations:list",
            kwargs={
                "wname": collection.workspace.name,
                "ccat": collection.category,
                "cname": collection.name,
            },
        )

        return ResourcePlace(
            title=format_html(
                "Relations of <code>{collection}</code> collection",
                collection=collection,
            ),
            url=url,
            breadcrumb="Relations",
            parent=collection.ui(self.request).place,
        )

    @cached_property
    def place_collection_relation_create(self) -> ResourcePlace:
        """Return a place to create a CollectionRelation."""
        collection = self.instance

        return ResourcePlace(
            title=format_html(
                "Create relation for <code>{collection}</code>",
                collection=collection,
            ),
            breadcrumb="Create",
            url=reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": collection.workspace.name,
                    "ccat": collection.category,
                    "cname": collection.name,
                },
            ),
            parent=self.place_collection_relation_list,
        )

    @cached_property
    def place_collection_relation_reorder_requires(self) -> ResourcePlace:
        """Return a place for re-ordering Requires relations."""
        collection = self.instance

        return ResourcePlace(
            title=format_html(
                "Reorder <code>{requires}</code> for <code>{collection}</code>",
                requires=CollectionRelation.Types.SUITE_REQUIRES.label,
                collection=collection,
            ),
            breadcrumb="Reorder",
            url=reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": collection.workspace.name,
                    "ccat": collection.category,
                    "cname": collection.name,
                },
            ),
            parent=self.place,
        )

    @cached_property
    def place_suite_root(self) -> ResourcePlace:
        """Return a place for showing an archive."""
        suite = self.instance
        return ResourcePlace(
            title=format_html(
                "Suite <code>{collection}</code>", collection=suite
            ),
            breadcrumb=str(suite),
            url=reverse(
                "suite-root",
                kwargs={
                    "scope": suite.workspace.scope.name,
                    "workspace": suite.workspace.name,
                    "suite": suite.name,
                },
            ),
            parent=suite.workspace.ui(self.request).place_suite_list,
        )


class CollectionItemUI(UI[CollectionItem]):
    """UI helpers for CollectionItem instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this collection item."""
        return ResourcePlace(
            title=self.instance.name,
            breadcrumb=self.instance.name,
            url=self.instance.get_absolute_url(),
            parent=self.instance.parent_collection.ui(self.request).place,
        )

    @override
    def build_ui_shortcuts(self) -> list[ui_shortcuts.UIShortcut]:
        result = super().build_ui_shortcuts()
        result.append(ui_shortcuts.create_collection_item(self.instance))
        if artifact := self.instance.artifact:
            result += [
                ui_shortcuts.create_artifact_view(artifact),
                ui_shortcuts.create_artifact_download(artifact),
            ]
        return result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for pools of workers."""

import logging
from functools import cached_property
from typing import TYPE_CHECKING, override

import pydantic
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import JSONField, QuerySet
from django.urls import reverse
from django.utils import timezone

from debusine.db.models.scopes import Scope
from debusine.server.worker_pools import (
    ScopeWorkerPoolLimits,
    WorkerPoolInterface,
    WorkerPoolLimits,
    WorkerPoolSpecifications,
    provider_interface,
    worker_pool_specifications_model,
)

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.workers import Worker
    from debusine.web.views.ui.worker_pools import WorkerPoolUI
else:
    TypedModelMeta = object

logger = logging.getLogger(__name__)


class WorkerPoolManager(models.Manager["WorkerPool"]):
    """Manager for WorkerPool model."""

    def enabled(self) -> QuerySet["WorkerPool"]:
        """Return connected workers."""
        return WorkerPool.objects.filter(enabled=True)


class WorkerPool(models.Model):
    """Database model of a worker pool."""

    name = models.SlugField(
        unique=True,
        help_text='Human readable name of the worker pool',
    )
    provider_account = models.ForeignKey("Asset", on_delete=models.PROTECT)
    enabled = models.BooleanField(default=True)
    architectures = JSONField(default=list)
    tags = JSONField(default=list, blank=True)
    specifications = JSONField(default=dict)
    instance_wide = models.BooleanField(default=True)
    ephemeral = models.BooleanField(default=False)
    limits = JSONField(default=dict, blank=True)
    apt_authentication = models.ForeignKey(
        "Asset",
        on_delete=models.PROTECT,
        related_name="+",
        null=True,
        blank=True,
    )

    registered_at = models.DateTimeField()

    objects = WorkerPoolManager()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"

    @override
    def __str__(self) -> str:
        """Return the id and name of the WorkerPool."""
        return f"Id: {self.id} Name: {self.name}"

    def get_absolute_url(self) -> str:
        """Return an absolute URL to view this WorkerPool."""
        return reverse("worker-pools:detail", kwargs={"name": self.name})

    def ui(self, request: "HttpRequest") -> "WorkerPoolUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.worker_pools import WorkerPoolUI

        return WorkerPoolUI.get(request, self)

    @property
    def limits_model(self) -> WorkerPoolLimits:
        """Return the pydantic model for limits."""
        return WorkerPoolLimits.model_validate(self.limits)

    @property
    def specifications_model(self) -> WorkerPoolSpecifications:
        """Return the pydantic model for specifications."""
        model = worker_pool_specifications_model(self.specifications)
        provider_model = self.provider_account.data_model
        assert hasattr(provider_model, "provider_type")
        if provider_model.provider_type != model.provider_type:
            raise ValueError(
                f"specifications for worker_pool {self.name} do not have a "
                f"provider_account with a matching provider_type."
            )
        return model

    @property
    def workers_running(self) -> QuerySet["Worker"]:
        """Return the Worker instances that are currently running."""
        return self.worker_set.filter(instance_created_at__isnull=False)

    @property
    def workers_stopped(self) -> QuerySet["Worker"]:
        """Return the Worker instances that are currently stopped."""
        return self.worker_set.filter(instance_created_at__isnull=True)

    @cached_property
    def provider_interface(self) -> WorkerPoolInterface:
        """Return a WorkerPoolInterface instance for this pool."""
        return provider_interface(self)

    def launch_workers(
        self, count: int, override_disabled: bool = False
    ) -> None:
        """Launch count additional worker instances."""
        from debusine.db.models.auth import Token
        from debusine.db.models.workers import Worker

        if not (self.enabled or override_disabled):
            raise ValueError("Pool is disabled, refusing to launch workers.")

        available = self.workers_stopped.count()
        if available < count:
            Worker.objects.create_pool_members(self, count - available)

        launched = 0
        while launched < count:
            with transaction.atomic():
                worker = (
                    self.workers_stopped.select_for_update(skip_locked=True)
                    .order_by("name")
                    .first()
                )
                if worker is None:  # pragma: no cover
                    # Locked or deleted since we expanded the pool, above
                    return

                old_activation_token = worker.activation_token
                worker.activation_token = (
                    Token.objects.create_worker_activation()
                )
                worker.save()
                if old_activation_token is not None:
                    old_activation_token.delete()
                if worker.token is not None:
                    worker.token.disable()

                self.provider_interface.launch_worker(worker)
                launched += 1

    def terminate_worker(self, worker: "Worker") -> None:
        """Terminate the specified worker instance."""
        # Import here to prevent circular imports
        from debusine.db.context import context
        from debusine.db.models.work_requests import (
            CannotRetry,
            WorkRequest,
            WorkRequestRetryReason,
        )

        if worker.worker_pool != self:
            raise ValueError(
                f"pool {self} cannot terminate worker"
                f" for pool {worker.worker_pool}"
            )

        # Commit this early to avoid scheduling any more work on the worker
        with transaction.atomic():
            if worker.activation_token is not None:
                worker.activation_token.disable()
            if worker.token is not None:
                worker.token.disable()

        if worker.instance_created_at is not None:
            WorkerPoolStatistics.objects.create(
                worker_pool=self,
                worker=worker,
                runtime=int(
                    (
                        timezone.now() - worker.instance_created_at
                    ).total_seconds()
                ),
            )

        # Trigger worker termination
        with transaction.atomic():
            self.provider_interface.terminate_worker(worker)

        with transaction.atomic():
            # De-assign any pending tasks assigned to the worker
            for pending in WorkRequest.objects.reassignable(worker=worker):
                pending.de_assign_worker()

            # Retry any work requests that were previously running on the worker
            for running in WorkRequest.objects.running(worker=worker):
                running.mark_aborted()
                try:
                    with context.disable_permission_checks():
                        running.retry(
                            reason=WorkRequestRetryReason.WORKER_FAILED
                        )
                except CannotRetry as e:
                    logger.debug(  # noqa: G200
                        "Cannot retry previously-running work request: %s", e
                    )

    @override
    def clean(self) -> None:
        """
        Ensure that data is valid for this worker pool.

        :raise ValidationError: for invalid data.
        """
        try:
            self.limits_model
            self.specifications_model
        except pydantic.ValidationError as e:
            raise ValidationError(message=str(e)) from e


class ScopeWorkerPool(models.Model):
    """Through table for linking a WorkerPool to a Scope."""

    worker_pool = models.ForeignKey(WorkerPool, on_delete=models.CASCADE)
    scope = models.ForeignKey(Scope, on_delete=models.CASCADE)

    priority = models.IntegerField(default=0)
    limits = JSONField(default=dict, blank=True)

    @override
    def __str__(self) -> str:
        """Return the id and name of the ScopeWorkerPool."""
        return (
            f"Id: {self.id} WorkerPool: {self.worker_pool.name} "
            f"Scope: {self.scope.name}"
        )

    @property
    def limits_model(self) -> ScopeWorkerPoolLimits:
        """Return the pydantic model for limits."""
        return ScopeWorkerPoolLimits.model_validate(self.limits)


class WorkerPoolTaskExecutionStatistics(models.Model):
    """Time spent executing tasks in a scope, stored at completion."""

    worker_pool = models.ForeignKey(WorkerPool, on_delete=models.CASCADE)
    worker = models.ForeignKey(
        "Worker", null=True, blank=True, on_delete=models.SET_NULL
    )
    scope = models.ForeignKey(Scope, on_delete=models.CASCADE)
    timestamp = models.DateTimeField(auto_now_add=True)
    runtime = models.IntegerField()

    class Meta(TypedModelMeta):
        indexes = [
            models.Index(
                "timestamp", name="%(app_label)s_worker_pool_exec_ts_idx"
            ),
        ]


class WorkerPoolStatistics(models.Model):
    """Running time for historical worker instances, stored at shutdown."""

    worker_pool = models.ForeignKey(WorkerPool, on_delete=models.CASCADE)
    worker = models.ForeignKey(
        "Worker", null=True, blank=True, on_delete=models.SET_NULL
    )
    timestamp = models.DateTimeField(auto_now_add=True)
    runtime = models.IntegerField()

    class Meta(TypedModelMeta):
        indexes = [
            models.Index(
                "timestamp", name="%(app_label)s_worker_pool_stat_ts_idx"
            ),
        ]
